import typing
from pathlib import Path
from typing import List, Optional, Union

import solara
from solara.alias import rw

if typing.TYPE_CHECKING:
    import numpy as np


@solara.component
def Image(image: Union[str, Path, "np.ndarray"], width: Optional[str] = None, classes: List[str] = []):
    """Displays an image from a URL, Path or numpy data.

    ## Arguments
     * `image`: URL, Path or numpy data.
     * `width`: Width of the image, by default (None) the image is displayed at its original size.
        Other options are: "100%" (to full its parent size), "100px" (to display it at 100 pixels wide),
        or any other CSS width value.
     * `classes`: List of CSS classes to add to the image.
    """
    layout = {}
    if width:
        layout["width"] = width
    classes = ["solara-image"] + classes
    if isinstance(image, (str, Path)):
        path = Path(image)
        if path.exists():
            import ipywidgets

            format = ipywidgets.Image._guess_format("image", str(path))
            value = ipywidgets.Image._load_file_value(str(path))
            return rw.Image(
                value=value,
                format=format,
                _dom_classes=classes,  # type: ignore
                layout=layout,
            )
        elif isinstance(image, str):
            url_data = image.encode("utf8")
            return rw.Image(
                value=url_data,
                format="url",
                _dom_classes=classes,  # type: ignore
                layout=layout,
            )
    elif solara.util.isinstanceof(image, "numpy:ndarray"):
        value = solara.util.numpy_to_image(image, format="png")
        return rw.Image(
            value=value,
            format="image/png",
            _dom_classes=classes,  # type: ignore
            layout=layout,
        )
    raise TypeError(f"Only support URL, path or numpy array, not {image}")



from typing import Callable, List, Optional, TypeVar, cast

import ipyvuetify as v
import reacton

import solara
from solara.alias import rv

T = TypeVar("T")


def _get_button_value(button: reacton.core.Element):
    value = button.kwargs.get("value")
    if value is None:
        value = button.kwargs.get("label")
    if value is None and button.args:
        value = button.args[0]
    return value


@solara.value_component(None)
def ToggleButtonsSingle(
    value: Optional[T] = None, values: List[T] = [], children: List[reacton.core.Element] = [], on_value: Optional[Callable[[T], None]] = None
) -> reacton.core.ValueElement[v.BtnToggle, T]:
    """ToggleButtons for selecting a single value.

    ## Arguments

    * `value`: The currently selected value.
    * `values`: List of values to select from.
    * `children`: List of buttons to use as values.
    * `on_value`: Callback to call when the value changes.
    """
    children = [solara.Button(label=str(value)) for value in values] + children
    values = values + [_get_button_value(button) for button in children]  # type: ignore
    index, set_index = solara.use_state_or_update(values.index(value) if value is not None else 0, key="index")

    def on_index(index):
        set_index(index)
        value = values[index]
        if on_value:
            on_value(value)

    return cast(reacton.core.ValueElement[v.BtnToggle, T], rv.BtnToggle(children=children, multiple=False, mandatory=True, v_model=index, on_v_model=on_index))


@solara.value_component(None)
def ToggleButtonsMultiple(
    value: List[T], values: List[T] = [], children: List[reacton.core.Element] = [], on_value: Callable[[List[T]], None] = None
) -> reacton.core.ValueElement[v.BtnToggle, List[T]]:
    """ToggleButtons for selecting multiple values.

    ## Arguments

    * `value`: The currently selected values.
    * `values`: List of values to select from.
    * `children`: List of buttons to use as values.
    * `on_value`: Callback to call when the value changes.
    """
    children = [solara.Button(label=str(value)) for value in values] + children
    allvalues = values + [_get_button_value(button) for button in children]
    indices, set_indices = solara.use_state_or_update([allvalues.index(k) for k in value], key="index")

    def on_indices(indices):
        set_indices(indices)
        value = [allvalues[k] for k in indices]
        if on_value:
            on_value(value)

    return cast(
        reacton.core.ValueElement[v.BtnToggle, List[T]], rv.BtnToggle(children=children, multiple=True, mandatory=False, v_model=indices, on_v_model=on_indices)
    )



import abc
import logging
from typing import Callable, List, Optional, Tuple, Union, cast

import solara
from solara import _using_solara_server

logger = logging.getLogger("solara.router")


class _LocationBase(abc.ABC):
    @property
    def pathname(self):
        pass

    @pathname.setter
    # mypy does not accept this
    # @abc.abstractmethod
    def pathname(self):
        pass


class _Location(_LocationBase):
    def __init__(self, pathname, setter: Callable[[str], None]) -> None:
        self._pathname = pathname
        self.setter = setter

    @property
    def pathname(self):
        return self._pathname

    @pathname.setter
    def pathname(self, value):
        # import pdb

        # pdb.set_trace()
        self._pathname = value
        self.setter(self._pathname)


class Router:
    search: Optional[str]

    def __init__(self, path: str, routes: List[solara.Route], set_path: Callable[[str], None] = None):
        # see https://developer.mozilla.org/en-US/docs/Web/API/Location for anatomy/nomenclature
        if "?" in path:
            self.path, self.search = path.split("?", 1)
        else:
            self.path = path
            self.search = None
        del path
        self.set_path = set_path
        self.parts = (self.path or "").strip("/").split("/")
        self.routes = routes
        self.root_path = ""
        if _using_solara_server():
            import solara.server.settings

            self.root_path = solara.server.settings.main.root_path or ""
        # each route in this list corresponds to a part in self.parts
        self.path_routes: List["solara.Route"] = []
        self.path_routes_siblings: List[List["solara.Route"]] = []  # siblings including itself
        # routes = routes.copy()
        route = None
        for part in self.parts:
            for route in routes:
                if (route.path == part) or (route.path == "/" and not part):
                    self.path_routes.append(route)
                    self.path_routes_siblings.append(routes)
                    routes = route.children
                    break
        if len(self.parts) == len(self.path_routes):
            # e.g. '/foo/bar' -> ['foo', 'bar'] and bar has a default route
            # but if '' -> [''] we should not
            route = self.path_routes[-1]
            if route:
                default_routes = [k for k in route.children if k.path == "/"]
                if self.parts and self.parts[0] and default_routes:
                    self.path_routes.append(default_routes[0])
                    self.path_routes_siblings.append(route.children)

        assert len(self.path_routes) == len(self.path_routes_siblings)
        self.possible_match = (len(self.path_routes[-1].children) == 0) if self.path_routes else False

    def push(self, path: str):
        assert self.set_path is not None
        self.set_path(path)


router_context = solara.create_context(Router("", []))
_location_context = solara.create_context(cast(_LocationBase, _Location("", lambda x: None)))

route_level_context = solara.create_context(0)


def use_route_level():
    route_level = solara.use_context(route_level_context)
    return route_level


def use_router() -> Router:
    return solara.use_context(router_context)


def use_route(level=0) -> Tuple[Optional[solara.Route], List[solara.Route]]:
    router = solara.use_context(router_context)
    route_level = solara.use_context(route_level_context)
    route_level_context.provide(route_level + 1)
    route_level += level
    if route_level < len(router.path_routes):
        return router.path_routes[route_level], router.path_routes_siblings[route_level]
    else:
        return None, []


def find_route(path: str) -> Optional[solara.Route]:
    router = solara.use_context(router_context)
    route_level = min(solara.use_context(route_level_context), len(router.path_routes_siblings) - 1)
    for route in router.path_routes_siblings[route_level]:
        if path.startswith(route.path) or (not path and route.path == "/"):
            return route
    return None


def use_pathname():
    location_proxy = solara.use_context(_location_context)

    def setter(value):
        location_proxy.pathname = value

    return location_proxy.pathname, setter


def resolve_path(path_or_route: Union[str, solara.Route], level=0) -> str:
    """Resolve a relative path or a route to an absolute path.

    If the path is a string and starts with a `/'`, it is returned as is.


    ## Typical usage:

    ```python
    ...
    route_current, routes_current_level = solara.routes()
    # route_current.path == "banana"
    path = solara.resolve_path(route_current)
    # path == "/fruit/banana"
    path_same = solara.resolve_path("banana")
    # path_same == path == "/fruit/banana"
    ...
    ```

    ## Arguments

     * path_or_route: a path string or a [`solara.Route`](/api/route) object to resolve.

    ## See also

     * [Multipage](/docs/howto/multipage).
     * [Understanding Routing](/docs/understanding/routing).


    """
    router = solara.use_context(router_context)
    if isinstance(path_or_route, str):
        path = path_or_route
        if path.startswith("/"):
            return path
        route_level = solara.use_context(route_level_context) + level - 1
        parts = [*router.parts[:route_level], path]
        path = "/" + "/".join(parts)
        if path.startswith("//"):
            path = path[1:]
        return path
    elif isinstance(path_or_route, solara.Route):
        route: solara.Route = path_or_route
        path = _resolve_path("/", route, router.routes)
        if path.startswith("//"):
            path = path[1:]
        return path


def _resolve_path(prefix: str, findroute: solara.Route, routes: List[solara.Route]):
    for route in routes:
        path = (prefix + "/" + route.path) if route.path != "/" else prefix
        if findroute is route:
            return path
        possible_path = _resolve_path(path, findroute=findroute, routes=route.children)
        if possible_path is not None:
            return possible_path



import numpy as np
import reacton.bqplot as bqplot
import reacton.ipywidgets as w

import solara

x0 = np.linspace(0, 2, 100)

title = "Lines graph using bqplot"


@solara.component
def Page(x=x0, ymax=5):
    with w.VBox() as main:
        y = x ** w.slider_float(1, min=0.1, max=3, description="Exponent")

        x_scale = bqplot.LinearScale()
        if w.checkbox(False, description="Log scale"):
            y_scale = bqplot.LogScale(min=0.1, max=ymax)
        else:
            y_scale = bqplot.LinearScale(min=0, max=ymax)

        display_legend = w.checkbox(True, "Show legend")
        label = ""
        if display_legend:
            label = w.text("My Label", "legend label")

        color = w.color("red", "Line color")

        lines = bqplot.Lines(x=x, y=y, scales={"x": x_scale, "y": y_scale}, stroke_width=3, colors=[color], display_legend=display_legend, labels=[label])
        x_axis = bqplot.Axis(scale=x_scale)
        y_axis = bqplot.Axis(scale=y_scale, orientation="vertical")
        bqplot.Figure(axes=[x_axis, y_axis], marks=[lines], scale_x=x_scale, scale_y=y_scale, layout={"min_width": "800px"})

    return main



import hashlib
import logging
import textwrap
from typing import Any, Dict, List, Union

import ipyvuetify as v
import pygments
import pymdownx.highlight
import pymdownx.superfences
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name

import solara
import solara.components.applayout

logger = logging.getLogger(__name__)

html_no_execute_enabled = "<div><i>Solara execution is not enabled</i></div>"


def _run_solara(code):
    ast = compile(code, "markdown", "exec")
    local_scope: Dict[Any, Any] = {}
    exec(ast, local_scope)
    app = None
    if "app" in local_scope:
        app = local_scope["app"]
    elif "Page" in local_scope:
        Page = local_scope["Page"]
        app = solara.components.applayout._AppLayoutEmbed(children=[Page()])
    else:
        raise NameError("No Page of app defined")
    box = v.Html(tag="div")
    box, rc = solara.render(app, container=box)
    widget_id = box._model_id
    return (
        '<div class="solara-markdown-output v-card v-sheet elevation-7">'
        f'<jupyter-widget widget="IPY_MODEL_{widget_id}">loading widget...</jupyter-widget>'
        '<div class="v-messages">Live output</div></div>'
    )


def _markdown_template(html, style=""):
    return (
        """
<template>
    <div class="solara-markdown rendered_html jp-RenderedHTMLCommon" style=\""""
        + style
        + """\">"""
        + html
        + """</div>
</template>

<script>
module.exports = {
    mounted() {
        mermaid.init()
        this.$el.querySelectorAll("a").forEach(a => this.setupRouter(a))
        window.md = this.$el
    },
    methods: {
        setupRouter(a) {
            let href = a.attributes['href'].value;
            if(href.startsWith("./")) {
                // TODO: should we really do this?
                href = location.pathname + href.substr(1);
                a.attributes['href'].href = href;
                // console.log("change href to", href);
            }
            if(href.startsWith("./") || href.startsWith("/")) {
                // console.log("connect link with href=", href, "to router")
                a.onclick = e => {
                    console.log("clicked", href)
                    if(href.startsWith("./")) {
                        solara.router.push(href);
                    } else {
                        solara.router.push(href);
                    }
                    e.preventDefault()
                }
            } else if(href.startsWith("#")) {
                // console.log("connect anchor with href=", href, "to custom javascript due to using <base>")
                href = location.pathname + href;
                a.attributes['href'].value = href;
            } else {
                console.log("href", href, "is not a local link")
            }
        }
    },
    updated() {
        // if the html gets update, re-run mermaid
        mermaid.init()
    }
}
</script>
    """
    )


def _highlight(src, language, unsafe_solara_execute, extra, *args, **kwargs):
    """Highlight a block of code"""

    run_src_with_solara = False
    if language == "solara":
        run_src_with_solara = True
        language = "python"

    lexer = get_lexer_by_name(language)
    formatter = HtmlFormatter()
    html = pygments.highlight(src, lexer, formatter)

    if run_src_with_solara:
        if unsafe_solara_execute:
            html_widget = _run_solara(src)
            return html + html_widget
        else:
            return html + html_no_execute_enabled
    else:
        return html


@solara.component
def MarkdownIt(md_text: str, highlight: List[int] = [], unsafe_solara_execute: bool = False):
    md_text = textwrap.dedent(md_text)

    from markdown_it import MarkdownIt as MarkdownItMod
    from mdit_py_plugins import container, deflist  # noqa: F401
    from mdit_py_plugins.footnote import footnote_plugin  # noqa: F401
    from mdit_py_plugins.front_matter import front_matter_plugin  # noqa: F401

    def highlight_code(code, name, attrs):
        return _highlight(code, name, unsafe_solara_execute, attrs)

    md = MarkdownItMod(
        "js-default",
        {
            "html": True,
            "typographer": True,
            "highlight": highlight_code,
        },
    )
    md = md.use(container.container_plugin, name="note")
    html = md.render(md_text)
    hash = hashlib.sha256((html + str(unsafe_solara_execute) + repr(highlight)).encode("utf-8")).hexdigest()
    return v.VuetifyTemplate.element(template=_markdown_template(html)).key(hash)


@solara.component
def Markdown(md_text: str, unsafe_solara_execute=False, style: Union[str, Dict, None] = None):
    """Renders markdown text

    Renders markdown using https://python-markdown.github.io/

    ## Arguments

     * `md_text`: The markdown text to render
     * `unsafe_solara_execute`: If True, code marked with language "solara" will be executed. This is potentially unsafe
        if the markdown text can come from user input and should only be used for trusted markdown.
     * `style`: A string or dict of css styles to apply to the rendered markdown.

    """
    import markdown

    md_text = textwrap.dedent(md_text)
    style = solara.util._flatten_style(style)

    def make_markdown_object():
        def highlight(src, language, *args, **kwargs):
            try:
                return _highlight(src, language, unsafe_solara_execute, *args, **kwargs)
            except Exception as e:
                logger.exception("Error highlighting code: %s", src)
                return repr(e)

        return markdown.Markdown(  # type: ignore
            extensions=[
                "pymdownx.highlight",
                "pymdownx.superfences",
                "pymdownx.emoji",
                "toc",  # so we get anchors for h1 h2 etc
            ],
            extension_configs={
                "pymdownx.superfences": {
                    "custom_fences": [
                        {
                            "name": "mermaid",
                            "class": "mermaid",
                            "format": pymdownx.superfences.fence_div_format,
                        },
                        {
                            "name": "solara",
                            "class": "",
                            "format": highlight,
                        },
                    ],
                },
            },
        )

    md = solara.use_memo(make_markdown_object, dependencies=[unsafe_solara_execute])
    html = md.convert(md_text)
    # if we update the template value, the whole vue tree will rerender (ipvue/ipyvuetify issue)
    # however, using the hash we simply generate a new widget each time
    hash = hashlib.sha256((html + str(unsafe_solara_execute)).encode("utf-8")).hexdigest()
    return v.VuetifyTemplate.element(template=_markdown_template(html, style)).key(hash)



import solara


@solara.component
def Page():
    clicks, set_clicks = solara.use_state(0)

    color = "green"
    if clicks >= 5:
        color = "red"

    def on_click():
        set_clicks(clicks + 1)
        print("clicks", clicks)  # noqa

    solara.Button(label=f"Clicked: {clicks}", on_click=on_click, color=color)



from typing import Callable, List, Optional, TypeVar, cast

import ipyvuetify as v
import reacton.core

import solara
from solara.alias import rv

T = TypeVar("T")


@solara.value_component(None)
def Select(label: str, values: List[T], value: Optional[T] = None, on_value: Callable[[T], None] = None) -> reacton.core.ValueElement[v.Select, T]:
    """Select a single value from a list of values.

    ## Arguments

     * `label`: Label to display next to the select.
     * `value`: The currently selected value.
     * `values`: List of values to select from.
     * `on_value`: Callback to call when the value changes.

    """
    return cast(
        reacton.core.ValueElement[v.Select, T],
        rv.Select(
            v_model=value,
            on_v_model=on_value,
            items=values,
            label=label,
            dense=True,
        ),
    )


@solara.value_component(None)
def SelectMultiple(
    label: str,
    values: List[T],
    all_values: List[T],
    on_value: Callable[[List[T]], None] = None,
) -> reacton.core.ValueElement[v.Select, List[T]]:
    """Select multiple values from a list of values.

    ## Arguments

        * `label`: Label to display next to the select.
        * `values`: List of currently selected values.
        * `all_values`: List of all values to select from.
        * `on_value`: Callback to call when the value changes.
    """
    return cast(
        reacton.core.ValueElement[v.Select, List[T]],
        rv.Select(
            v_model=values,
            on_v_model=on_value,
            items=all_values,
            label=label,
            multiple=True,
            dense=False,
        ),
    )



import base64
import contextlib
import os
import sys
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import PIL.Image

import solara


def github_url(file):
    rel_path = os.path.relpath(file, Path(solara.__file__).parent.parent)
    github_url = solara.github_url + f"/blob/{solara.git_branch}/" + rel_path
    return github_url


def load_file_as_data_url(file_name, mime):
    with open(file_name, "rb") as f:
        data = f.read()
    return f"data:{mime};base64," + base64.b64encode(data).decode("utf-8")


def isinstanceof(object, spec: str):
    """Check if object is instance of type '<modulename>:<classname>'

    This can avoid a runtime dependency, since we do not need to import `modulename`.

    >>> import numpy as np
    >>> isinstanceof(np.arange(2), "numpy:ndarray")
    True
    """
    module_name, classname = spec.split(":")
    module = sys.modules.get(module_name)
    if module:
        cls = getattr(module, classname)
        return isinstance(object, cls)
    return False


def numpy_to_image(data: "np.ndarray", format="png"):
    import io

    if data.ndim == 3:
        if data.shape[2] == 3:
            im = PIL.Image.fromarray(data[::], "RGB")
        elif data.shape[2] == 4:
            im = PIL.Image.fromarray(data[::], "RGBA")
        else:
            raise ValueError(f"Expected last dimension to have 3 or 4 dimensions, total shape we got was {data.shape}")
        f = io.BytesIO()
        im.save(f, format)
        return f.getvalue()
    else:
        raise ValueError(f"Expected an image with 3 dimensions (height, width, channe), not {data.shape}")


@contextlib.contextmanager
def cwd(path):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


def numpy_equals(a, b):
    if a is b:
        return True
    if a is None or b is None:
        return False
    if np.all(a == b):
        return True
    return False


def _combine_classes(class_list: List[str]) -> str:
    return " ".join(class_list)


def _flatten_style(style: Union[str, Dict, None] = None) -> str:
    if style is None:
        return ""
    elif isinstance(style, str):
        return style
    elif isinstance(style, dict):
        return ";".join(f"{k}:{v}" for k, v in style.items())
    else:
        raise ValueError(f"Expected style to be a string or dict, got {type(style)}")


def import_item(name: str):
    """Import an object by name like solara.cache.LRU"""
    parts = name.rsplit(".", 2)
    if len(parts) == 1:
        return __import__(name)
    else:
        module = __import__(".".join(parts[:-1]), fromlist=[parts[-1]])
        return getattr(module, parts[-1])


def get_solara_home() -> Path:
    """Get solara home directory, defaults to $HOME/.solara.

    The $SOLARA_HOME environment variable can be set to override this default.

    If both $SOLARA_HOME and $HOME are not defined, the current working directory is used.
    """
    if "SOLARA_HOME" in os.environ:
        return Path(os.environ["SOLARA_HOME"])
    elif "HOME" in os.environ:
        return Path(os.path.join(os.environ["HOME"], ".solara"))
    else:
        return Path(os.getcwd())


def parse_size(size: str) -> int:
    """Given a human readable size, return the number of bytes.

    Supports GB, MB, KB, and bytes. E.g. 10GB, 500MB, 1KB, 1000

    Commas and _ are ignored, e.g. 1,000,000 is the same as 1000000.
    """
    size = size.replace(",", "").replace("_", "").upper()
    if size.endswith("GB"):
        return int(float(size[:-2]) * 1024 * 1024 * 1024)
    elif size.endswith("MB"):
        return int(float(size[:-2]) * 1024 * 1024)
    elif size.endswith("KB"):
        return int(float(size[:-2]) * 1024)
    elif size.endswith("B"):
        return int(float(size[:-1]))
    else:
        return int(size)



from packages.assets.hatch_build import npm_pack


def test_npm_pack(tmp_path_factory):
    base_cache_dir = tmp_path_factory.mktemp("cdn")

    package = "@widgetti/solara-vuetify-app"
    version = "0.0.1-alpha.1"
    npm_pack(base_cache_dir, package, version)

    some_file = base_cache_dir / f"{package}@{version}" / "dist" / "037d830416495def72b7881024c14b7b.woff2"
    assert some_file.is_file()
    assert len(some_file.read_bytes()) == 15436



"""
# Sliders

To support proper typechecks, we have multiple slider (all wrapping the ipyvuetify sliders).


"""
import datetime

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    with solara.VBox() as main:
        with solara.Card("Integer"):
            int_value, set_int_value = solara.use_state(42)
            solara.SliderInt("Some integer", value=int_value, min=-10, max=120, on_value=set_int_value)
            solara.Markdown(f"**Int value**: {int_value}")

        with solara.Card("Integer range"):
            int_range_value, set_int_range_value = solara.use_state((0, 42))
            solara.SliderRangeInt("Some integer range", value=int_range_value, min=-10, max=120, on_value=set_int_range_value)
            solara.Markdown(f"**Int range value**: {int_range_value}")

        with solara.Card("Float"):
            float_value, set_float_value = solara.use_state(42.4)
            solara.SliderFloat("Some float", value=float_value, min=-10, max=120, on_value=set_float_value)
            solara.Markdown(f"**Float value**: {float_value}")

        with solara.Card("Float range"):
            float_range_value, set_float_range_value = solara.use_state((0.1, 42.4))
            solara.SliderRangeFloat("Some float", value=float_range_value, min=-10, max=120, on_value=set_float_range_value)
            solara.Markdown(f"**Float value**: {float_range_value}")

        with solara.Card("Values"):
            values = "Python C++ Java JavaScript TypeScript BASIC".split()
            value, set_value = solara.use_state(values[0])
            solara.SliderValue("Language", value, values=values, on_value=set_value)
            solara.Markdown(f"**Value**: {value}")

        with solara.Card("Dates"):
            date, set_date = solara.use_state(datetime.date(1981, 7, 28))
            solara.SliderDate("Some date", value=date, on_value=set_date)
            solara.Markdown(f"**Date**: {date.strftime('%Y-%b-%d')}")

    return main


__doc__ += "# SliderInt"
__doc__ += apidoc(solara.SliderInt.f)  # type: ignore
__doc__ += "# SliderRangeInt"
__doc__ += apidoc(solara.SliderRangeInt.f)  # type: ignore

__doc__ += "# SliderFloat"
__doc__ += apidoc(solara.SliderFloat.f)  # type: ignore
__doc__ += "# SliderRangeFloat"
__doc__ += apidoc(solara.SliderRangeFloat.f)  # type: ignore

__doc__ += "# SliderValue"
__doc__ += apidoc(solara.SliderValue.f)  # type: ignore

__doc__ += "# SliderDate"
__doc__ += apidoc(solara.SliderDate.f)  # type: ignore



import contextlib
import dataclasses
import functools
import inspect
import io
import json
import logging
import os
import sys

# import tempfile
import threading
import time
import urllib.request
import uuid
from typing import IO, Any, Callable, Iterator, Optional, Tuple, TypeVar, Union, cast

import reacton

import solara
from solara.datatypes import FileContentResult, Result, ResultState

SOLARA_ALLOW_OTHER_TRACER = os.environ.get("SOLARA_ALLOW_OTHER_TRACER", False) in (True, "True", "true", "1")

logger = logging.getLogger("react-ipywidgets.extra.hooks")
chunk_size_default = 1024**2

__all__ = [
    "use_thread",
    "use_download",
    "use_fetch",
    "use_json_load",
    "use_json",
    "use_file_content",
    "use_uuid4",
    "use_unique_key",
    "use_state_or_update",
    "use_previous",
]
T = TypeVar("T")
U = TypeVar("U")

MaybeResult = Union[T, Result[T]]


# not available in python 3.6
class nullcontext(contextlib.AbstractContextManager):
    def __init__(self, enter_result=None):
        self.enter_result = enter_result

    def __enter__(self):
        return self.enter_result

    def __exit__(self, *excinfo):
        pass


# inherit from BaseException so less change of being caught
# in an except
class CancelledError(BaseException):
    pass


def use_retry(*actions: Callable[[], Any]):
    counter, set_counter = solara.use_state(0)

    def retry():
        for action in actions:
            action()
        set_counter(lambda counter: counter + 1)

    return counter, retry


def use_thread(
    callback=Union[
        Callable[[threading.Event], T],
        Iterator[Callable[[threading.Event], T]],
        Callable[[], T],
        Iterator[Callable[[], T]],
    ],
    dependencies=[],
    intrusive_cancel=True,
) -> Result[T]:
    def make_event(*_ignore_dependencies):
        return threading.Event()

    def make_lock():
        return threading.Lock()

    lock: threading.Lock = solara.use_memo(make_lock, [])
    updater = use_force_update()
    result_state, set_result_state = solara.use_state(ResultState.INITIAL)
    error = solara.use_ref(cast(Optional[Exception], None))
    result = solara.use_ref(cast(Optional[T], None))
    running_thread = solara.use_ref(cast(Optional[threading.Thread], None))
    counter, retry = use_retry()
    cancel: threading.Event = solara.use_memo(make_event, [*dependencies, counter])

    @contextlib.contextmanager
    def cancel_guard():
        if not intrusive_cancel:
            yield
            return

        def tracefunc(frame, event, arg):
            # this gets called at least for every line executed
            if cancel.is_set():
                rc = reacton.core._get_render_context(required=False)
                # we do not want to cancel the rendering cycle
                if rc is None or not rc._is_rendering:
                    # this will bubble up
                    raise CancelledError()
            if prev and SOLARA_ALLOW_OTHER_TRACER:
                prev(frame, event, arg)
            # keep tracing:
            return tracefunc

        # see https://docs.python.org/3/library/sys.html#sys.settrace
        # it is for the calling thread only
        # not every Python implementation has it
        prev = None
        if hasattr(sys, "gettrace"):
            prev = sys.gettrace()
        if hasattr(sys, "settrace"):
            sys.settrace(tracefunc)
        try:
            yield
        finally:
            if hasattr(sys, "settrace"):
                sys.settrace(prev)

    def run():
        set_result_state(ResultState.STARTING)

        def runner():
            wait_for_thread = None
            with lock:
                # if there is a current thread already, we'll need
                # to wait for it. copy the ref, and set ourselves
                # as the current one
                if running_thread.current:
                    wait_for_thread = running_thread.current
                running_thread.current = threading.current_thread()
            if wait_for_thread is not None:
                set_result_state(ResultState.WAITING)
                # don't start before the previous is stopped
                try:
                    wait_for_thread.join()
                except:  # noqa
                    pass
                if threading.current_thread() != running_thread.current:
                    # in case a new thread was started that also was waiting for the previous
                    # thread to st stop, we can finish this
                    return
            # we previously set current to None, but if we do not do that, we can still render the old value
            # while we can still show a loading indicator using the .state
            # result.current = None
            set_result_state(ResultState.RUNNING)

            sig = inspect.signature(callback)
            if sig.parameters:
                f = functools.partial(callback, cancel)
            else:
                f = callback
            try:
                try:
                    # we only use the cancel_guard context manager around
                    # the function calls to f. We don't want to guard around
                    # a call to react, since that might slow down rendering
                    # during rendering
                    with cancel_guard():
                        value = f()
                    if inspect.isgenerator(value):
                        while True:
                            try:
                                with cancel_guard():
                                    result.current = next(value)
                                    error.current = None
                            except StopIteration:
                                break
                            # assigning to the ref doesn't trigger a rerender, so do it manually
                            updater()
                        if threading.current_thread() == running_thread.current:
                            set_result_state(ResultState.FINISHED)
                    else:
                        result.current = value
                        error.current = None
                        if threading.current_thread() == running_thread.current:
                            set_result_state(ResultState.FINISHED)
                except Exception as e:
                    error.current = e
                    if threading.current_thread() == running_thread.current:
                        logger.exception(e)
                        set_result_state(ResultState.ERROR)
                    return
                except CancelledError:
                    pass
                    # this means this thread is cancelled not be request, but because
                    # a new thread is running, we can ignore this
            finally:
                if threading.current_thread() == running_thread.current:
                    running_thread.current = None
                    logger.info("thread done!")
                    if cancel.is_set():
                        set_result_state(ResultState.CANCELLED)

        logger.info("starting thread: %r", runner)
        thread = threading.Thread(target=runner, daemon=True)
        thread.start()

        def cleanup():
            cancel.set()  # cleanup for use effect

        return cleanup

    solara.use_side_effect(run, dependencies + [counter])
    return Result[T](value=result.current, error=error.current, state=result_state, cancel=cancel.set, _retry=retry)


def use_download(
    f: MaybeResult[Union[str, os.PathLike, IO]], url, expected_size=None, delay=None, return_content=False, chunk_size=chunk_size_default
) -> Result:
    if not isinstance(f, Result):
        f = Result(value=f)
    assert isinstance(f, Result)
    content_length, set_content_length = solara.use_state(expected_size, key="content_length")
    downloaded_length = 0
    file_object = hasattr(f.value, "tell")
    if not file_object:
        file_path = cast(Union[str, os.PathLike], f.value)
        if os.path.exists(file_path) and expected_size is not None:
            file_size = os.path.getsize(file_path)
            if file_size == expected_size:
                downloaded_length = file_size
    downloaded_length, set_downloaded_length = solara.use_state(downloaded_length, key="downloaded_length")

    def download(cancel: threading.Event):
        assert isinstance(f, Result)
        nonlocal downloaded_length
        if expected_size is not None and downloaded_length == expected_size:
            return  # we already downloaded, but hooks cannot be conditional

        context: Any = None
        if file_object:
            context = nullcontext()
            output_file = cast(IO, f.value)
        else:
            # f = cast(Result[Union[str, os.PathLike]], f)
            output_file = context = open(f.value, "wb")  # type: ignore

        with context:
            with urllib.request.urlopen(url) as response:
                content_length = int(response.info()["Content-Length"])
                logger.info("content_length for %r = %r", url, content_length)
                set_content_length(content_length)
                bytes_read = 0
                while not cancel.is_set():
                    chunk = response.read(chunk_size)

                    if delay:
                        time.sleep(delay)
                    if not chunk:
                        break
                    bytes_read += len(chunk)
                    output_file.write(chunk)
                    set_downloaded_length(bytes_read)
        return bytes_read

    result: Result[Any] = use_thread(download, [f, url])
    # maybe we wanna check this
    # download_is_done = downloaded_length == content_length

    if content_length is not None:
        progress = downloaded_length / content_length
    else:
        progress = 0

    return dataclasses.replace(result, progress=progress)


def use_fetch(url, chunk_size=chunk_size_default):
    # re-use the same file like object
    f = solara.use_memo(io.BytesIO, [url])
    result = use_download(f, url, return_content=True, chunk_size=chunk_size)
    return dataclasses.replace(result, value=f.getvalue() if result.progress == 1 else None)


def compose_result(head, *tail):
    return head


def ensure_result(input: MaybeResult[T]) -> Result[T]:
    if isinstance(input, Result):
        return input
    else:
        return Result(value=input)


def make_use_thread(f: Callable[[T], U]):
    def use_result(input: MaybeResult[T]) -> Result[U]:
        input_result = ensure_result(input)

        def in_thread(cancel: threading.Event):
            if input_result.value:
                return f(input_result.value)

        return use_thread(in_thread, dependencies=[input_result.value])

    return use_result


@make_use_thread
def use_json_load(value: bytes):
    return json.loads(value)


def use_json(path):
    return use_fetch(path) | use_json_load


def use_file_content(path, watch=False) -> FileContentResult[bytes]:
    counter, retry = use_retry()

    def read_file(*ignore):
        try:
            with open(path, "rb") as f:
                return f.read()
        except Exception as e:
            return e

    result = None
    try:
        mtime = os.path.getmtime(path)
    except Exception:
        mtime = None

    content = solara.use_memo(read_file, dependencies=[path, mtime, counter])
    if result is not None:
        return result
    if isinstance(content, Exception):
        return FileContentResult[bytes](error=content, _retry=retry)
    else:
        return FileContentResult[bytes](value=content, _retry=retry)


def use_force_update() -> Callable[[], None]:
    """Returns a callable that can be used to force an update of a component.

    This is used when external state has change, and we need to re-render out component.
    """
    _counter, set_counter = solara.use_state(0, "force update counter")

    def updater():
        set_counter(lambda count: count + 1)

    return updater


def use_uuid4(dependencies=[]) -> str:
    """Generate a unique string using the uuid4 algorithm. Will only change when the dependencies change."""

    def make_uuid(*_ignore):
        return str(uuid.uuid4())

    return solara.use_memo(make_uuid, dependencies)


def use_unique_key(key: str = None, prefix: str = "", dependencies=[]) -> str:
    """Generate a unique string, or use key when not None. Dependencies are forwarded to `use_uuid4`."""
    uuid = use_uuid4(dependencies=dependencies)
    return prefix + (key or uuid)


def use_state_or_update(
    initial_or_updated: T, key: str = None, eq: Callable[[Any, Any], bool] = None
) -> Tuple[T, Callable[[Union[T, Callable[[T], T]]], None]]:
    """This is useful for situations where a prop can change from a parent
    component, which should be respected, and otherwise the internal
    state should be kept.
    """
    value, set_value = solara.use_state(initial_or_updated, key=key, eq=eq)

    def possibly_update():
        nonlocal value
        # only gets called when initial_or_updated changes
        set_value(initial_or_updated)
        # this make sure the return value gets updated directly
        value = initial_or_updated

    solara.use_memo(possibly_update, [initial_or_updated])
    return value, set_value


def use_previous(value: T, condition=True) -> T:
    ref = solara.use_ref(value)

    def assign():
        if condition:
            ref.current = value

    solara.use_effect(assign, [value])
    return ref.current



"""# Checkbox

"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    turbo_boost, set_turbo_boost = solara.use_state(True)
    with solara.VBox() as main:
        if turbo_boost:
            solara.Success("Turbo boost is on")
        else:
            solara.Warning("Turbo boost is off, you might want to turn it on")
        solara.Checkbox(label="Turbo boost", value=turbo_boost, on_value=set_turbo_boost)
    return main


__doc__ += apidoc(solara.Checkbox.f)  # type: ignore



import site
import sys
import uuid
from enum import Enum
from pathlib import Path
from typing import Optional

import pydantic
from filelock import FileLock

from .. import (  # noqa  # sidefx is that this module creates the ~/.solara directory
    settings,
)
from ..util import get_solara_home

if __file__.startswith(site.getuserbase()):
    prefix = site.getuserbase()
else:
    prefix = sys.prefix


class ThemeVariant(str, Enum):
    light = "light"
    dark = "dark"
    auto = "auto"


class ThemeSettings(pydantic.BaseSettings):
    variant: ThemeVariant = ThemeVariant.light
    variant_user_selectable: bool = True
    loader: str = "solara"

    class Config:
        env_prefix = "solara_theme_"
        case_sensitive = False
        env_file = ".env"


class SSG(pydantic.BaseSettings):
    # the first app create will initialize this if it is not set
    build_path: Optional[Path] = None
    enabled: bool = False
    headed: bool = False

    class Config:
        env_prefix = "solara_ssg_"
        case_sensitive = False
        env_file = ".env"


class Search(pydantic.BaseSettings):
    enabled: bool = False


class Telemetry(pydantic.BaseSettings):
    mixpanel_token: str = "91845eb13a68e3db4e58d64ad23673b7"
    mixpanel_enable: bool = True
    server_user_id: str = "not_set"
    server_fingerprint: str = str(uuid.getnode())
    server_session_id: str = str(uuid.uuid4())

    class Config:
        env_prefix = "solara_telemetry_"
        case_sensitive = False
        env_file = ".env"


class Assets(pydantic.BaseSettings):
    cdn: str = "https://cdn.jsdelivr.net/npm/"
    proxy: bool = True
    proxy_cache_dir: Path = Path(prefix + "/share/solara/cdn/")
    fontawesome_enabled: bool = True
    fontawesome_path: str = "/font-awesome@4.5.0/css/font-awesome.min.css"

    class Config:
        env_prefix = "solara_assets_"
        case_sensitive = False
        env_file = ".env"


class MainSettings(pydantic.BaseSettings):
    use_pdb: bool = False
    mode: str = "production"
    tracer: bool = False
    timing: bool = False
    root_path: Optional[str] = None

    class Config:
        env_prefix = "solara_"
        case_sensitive = False
        env_file = ".env"


main = MainSettings()
theme = ThemeSettings()
telemetry = Telemetry()
ssg = SSG()
search = Search()
assets = Assets()

assets.proxy_cache_dir.mkdir(exist_ok=True, parents=True)

if telemetry.server_user_id == "not_set":
    home = get_solara_home()
    server_user_id_file = home / "server_user_id.txt"
    with FileLock(str(server_user_id_file) + ".lock"):
        if not server_user_id_file.exists():
            server_user_id_file.write_text(str(uuid.uuid4()))
        telemetry.server_user_id = server_user_id_file.read_text()



import solara
from matplotlib.figure import Figure


@solara.component
def Page():
    # do this instead of plt.figure()
    fig = Figure()
    ax = fig.subplots()
    ax.plot([1, 2, 3], [1, 4, 9])
    return solara.FigureMatplotlib(fig)


def test_render():
    box, rc = solara.render(Page(), handle_error=False)
    assert len(box.children) == 1



import dataclasses
import importlib.util
import logging
import os
import pickle
import sys
import threading
import traceback
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, cast

import IPython.display
import ipywidgets as widgets
import reacton
from ipywidgets import DOMWidget, Widget
from reacton.core import Element, render

import solara

from ..util import cwd
from . import kernel, reload, settings, websocket
from .kernel import Kernel, WebsocketStreamWrapper
from .utils import nested_get, pdb_guard

WebSocket = Any
apps: Dict[str, "AppScript"] = {}
thread_lock = threading.Lock()

logger = logging.getLogger("solara.server.app")
state_directory = Path(".") / "states"
state_directory.mkdir(exist_ok=True)

reload.reloader.start()


class AppType(str, Enum):
    SCRIPT = "script"
    NOTEBOOK = "notebook"
    MODULE = "module"
    DIRECTORY = "directory"


def display(*args, **kwargs):
    print("display not implemented", args, kwargs)  # noqa


_warned_is_widget = False


def warn_is_widget():
    global _warned_is_widget
    if not _warned_is_widget:
        _warned_is_widget = True
        print(  # noqa
            "The app you are running is a widget, note that your app is run once before any user connected. "
            "If this causes a performance issue, please open an issue on GitHub."
        )


class AppScript:
    directory: Path
    routes: List[solara.Route]

    def __init__(self, name, default_app_name="Page"):
        self.fullname = name
        if reload.reloader.on_change:
            raise RuntimeError("Previous reloader still had a on_change attached, no cleanup?")
        reload.reloader.on_change = self.on_file_change

        self.app_name = default_app_name
        if ":" in self.fullname:
            self.name, self.app_name = self.fullname.rsplit(":", 1)
            if len(self.name) == 1:  # a windows drive letter, restore
                self.name = self.fullname
                self.app_name = default_app_name
        else:
            self.name = name
        self.path: Path = Path(self.name).resolve()
        try:
            context = get_current_context()
        except RuntimeError:
            context = None
        if context is not None:
            raise RuntimeError(f"We should not have an existing Solara app context when running an app for the first time: {context}")
        app_context = create_dummy_context()
        with app_context:
            app = self._execute()

        self._first_execute_app = app
        self._is_widget = isinstance(self._first_execute_app, widgets.Widget)
        app_context.close()
        if self._is_widget:
            warn_is_widget()

    def _execute(self):
        logger.info("Executing %s", self.name)
        app = None
        local_scope = {
            "display": IPython.display.display,
            "__name__": "__main__",
            "__file__": str(self.path),
        }
        ignore = list(local_scope)
        routes: Optional[List[solara.Route]] = None
        if self.path.is_dir():
            self.type = AppType.DIRECTORY
            # resolve the directory, because Path("file").parent.parent == "." != ".."
            self.directory = self.path.resolve()
            routes = solara.generate_routes_directory(self.path)
            app = solara.autorouting.RenderPage()
        elif self.name.endswith(".py"):
            self.type = AppType.SCRIPT
            # manually add the script to the watcher
            reload.reloader.watcher.add_file(self.path)
            self.directory = self.path.parent.resolve()
            with reload.reloader.watch():
                local_scope = {}
                with open(self.path) as f:
                    ast = compile(f.read(), self.path, "exec")
                    exec(ast, local_scope)
            app = nested_get(local_scope, self.app_name)
            routes = cast(Optional[List[solara.Route]], local_scope.get("routes"))
            if isinstance(app, Element):
                app = solara.AppLayout(children=[app])
            if isinstance(app, reacton.core.Component):
                app = solara.AppLayout(children=[app()])
        elif self.name.endswith(".ipynb"):
            self.type = AppType.NOTEBOOK
            # manually add the notebook to the watcher
            reload.reloader.watcher.add_file(self.path)
            self.directory = self.path.parent.resolve()
            import nbformat

            nb: nbformat.NotebookNode = nbformat.read(self.path, 4)
            with reload.reloader.watch(), cwd(Path(self.path).parent):
                for cell_index, cell in enumerate(nb.cells):
                    cell_index += 1  # used 1 based
                    if cell.cell_type == "code":
                        source = cell.source
                        cell_path = f"{self.path} input cell {cell_index}"
                        ast = compile(source, cell_path, "exec")
                        exec(ast, local_scope)
                app = nested_get(local_scope, self.app_name)
                routes = cast(Optional[List[solara.Route]], local_scope.get("routes"))
            if isinstance(app, Element):
                app = solara.AppLayout(children=[app])
            if isinstance(app, reacton.core.Component):
                app = solara.AppLayout(children=[app()])
        else:
            # the module itself will be added by reloader
            # automatically
            with reload.reloader.watch():
                self.type = AppType.MODULE
                try:
                    spec = importlib.util.find_spec(self.name)
                except ValueError:
                    if self.name not in sys.modules:
                        raise ImportError(f"Module {self.name} not found")
                    spec = importlib.util.spec_from_file_location(self.name, sys.modules[self.name].__file__)
                if spec is None:
                    raise ImportError(f"Module {self.name} cannot be found")
                assert spec is not None
                if spec.origin is None:
                    raise ImportError(f"Module {self.name} cannot be found, or is a namespace package")
                assert spec.origin is not None
                self.path = Path(spec.origin)
                self.directory = self.path.parent

                mod = importlib.import_module(self.name)

                local_scope = mod.__dict__
                if not hasattr(mod, "routes"):
                    if self.app_name == "Page":
                        routes = solara.generate_routes(mod)
                        app = solara.RenderPage()
                    else:
                        app = nested_get(local_scope, self.app_name)
                        routes = None
                else:
                    routes = mod.routes
                    app = nested_get(local_scope, self.app_name)
                    if app is None:
                        app = solara.autorouting.RenderPage()

        # this is not expected for modules, similar to `python script.py and python -m package.mymodule`
        if self.type in [AppType.SCRIPT, AppType.NOTEBOOK]:
            working_directory = str(self.path.parent)
            if working_directory not in sys.path:
                sys.path.insert(0, working_directory)

        if settings.ssg.build_path is None:
            settings.ssg.build_path = self.directory.parent.resolve() / "build"

        # auto enable search if search.json exists
        search_index_file = self.directory.parent / "assets" / "search.json"
        if search_index_file.exists():
            settings.search.enabled = True

        # this might be useful for development
        # but requires reloading of react in solara iself
        # for name, module in sys.modules.items():
        #     if name.startswith("reacton"):
        #         file = inspect.getfile(module)
        #         self.watcher.add_file(file)

        # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
        # we cannot set script name, because gunicorn uses it (and will crash)
        # os.environ["SCRIPT_NAME"] = self.name
        os.environ["PATH_TRANSLATED"] = str(self.path.resolve())

        if app is None:
            # workaround for backward compatibility
            app = local_scope.get("app")
        if app is None:
            app = local_scope.get("page")
        if app is None:
            import difflib

            options = [k for k in list(local_scope) if k not in ignore and not k.startswith("_")]
            matches = difflib.get_close_matches(self.app_name, options)
            msg = f"No object with name {self.app_name} found for {self.name} at {self.path}."
            if matches:
                msg += " Did you mean: " + " or ".join(map(repr, matches))
            else:
                msg += " We did find: " + " or ".join(map(repr, options))
            raise NameError(msg)
        if routes is None:
            self.routes = [solara.Route("/")]
        else:
            self.routes = routes

        # this might be useful for development
        # but requires reloading of react in solara iself
        # for name, module in sys.modules.items():
        #     if name.startswith("reacton"):
        #         file = inspect.getfile(module)
        #         self.watcher.add_file(file)

        # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
        # we cannot set script name, because gunicorn uses it (and will crash)
        # os.environ["SCRIPT_NAME"] = self.name
        os.environ["PATH_TRANSLATED"] = str(self.path.resolve())
        return app

    def close(self):
        reload.reloader.on_change = None
        context_values = list(contexts.values())
        contexts.clear()
        for context in context_values:
            context.close()

    def run(self):
        if self._is_widget:
            return self._execute()
        else:
            if reload.reloader.requires_reload:
                with thread_lock:
                    if reload.reloader.requires_reload:
                        self._first_execute_app = self._execute()
            return self._first_execute_app

    def on_file_change(self, name):
        path = Path(name)
        if path.suffix == ".vue":
            logger.info("Vue file changed: %s", name)
            template_content = path.read_text()
            for context in list(contexts.values()):
                with context:
                    for filepath, widget in context.templates.items():
                        if filepath == str(path):
                            widget.template = template_content
        else:
            logger.info("Reload requires due to change in module: %s", name)
            self.reload()

    def reload(self):
        # if multiple files change in a short time, we want to do this
        # not concurrently. Even better would be to do a debounce?
        with thread_lock:
            context_values = list(contexts.values())
            # save states into the context so the hot reload will
            # keep the same state
            for context in context_values:
                render_context = cast(reacton.core._RenderContext, context.app_object)
                if render_context:
                    with context:
                        # we save the state for when the app reruns, so we stay in the same state.
                        # (e.g. button clicks, chosen options etc)
                        # for instance a dataframe, needs to be pickled, because after the pandas
                        # module is reloaded, it's a different class type
                        logger.info("pickling state: %s", render_context.state_get())
                        try:
                            context.state = pickle.dumps(render_context.state_get())
                        except Exception as e:
                            logger.warning("Could not pickle state, next render the state will be lost: %s", e)
                        # clear/cleanup the render_context, so during reload we start
                        # from scratch
                        context.app_object = None
                        # we want to reuse the container
                        render_context.container = None
                        try:
                            render_context.close()
                        except Exception as e:
                            logger.exception("Could not close render context: %s", e)

            # ask all contexts/users to reload
            for context in context_values:
                with context:
                    context.reload()


@dataclasses.dataclass
class AppContext:
    id: str
    kernel: kernel.Kernel
    control_sockets: List[WebSocket] = dataclasses.field(default_factory=list)
    # this is the 'private' version of the normally global ipywidgets.Widgets.widget dict
    # see patch.py
    widgets: Dict[str, Widget] = dataclasses.field(default_factory=dict)
    # same, for ipyvue templates
    # see patch.py
    templates: Dict[str, Widget] = dataclasses.field(default_factory=dict)
    user_dicts: Dict[str, Dict] = dataclasses.field(default_factory=dict)
    # anything we need to attach to the context
    # e.g. for a react app the render context, so that we can store/restore the state
    app_object: Optional[Any] = None
    reload: Callable = lambda: None
    state: Any = None
    container: Optional[DOMWidget] = None

    def display(self, *args):
        print(args)  # noqa

    def __enter__(self):
        key = get_current_thread_key()
        current_context[key] = self

    def __exit__(self, *args):
        key = get_current_thread_key()
        current_context[key] = None

    def close(self):
        with self:
            if self.app_object is not None:
                if isinstance(self.app_object, reacton.core._RenderContext):
                    try:
                        self.app_object.close()
                    except Exception as e:
                        logger.exception("Could not close render context: %s", e)
                        # we want to continue, so we at least close all widgets
            import solara.server.patch

            assert isinstance(widgets.Widget.widgets, solara.server.patch.context_dict_widgets), f"Unexpected widget dict type: {type(widgets.Widget.widgets)}"
            assert widgets.Widget.widgets._get_context_dict() is self.widgets
            widgets.Widget.close_all()
            # what if we reference eachother
            # import gc
            # gc.collect()
        if self.id in contexts:
            del contexts[self.id]

    def _state_reset(self):
        path = state_directory / f"{self.id}.pickle"
        path = path.absolute()
        try:
            path.unlink()
        except:  # noqa
            pass
        del contexts[self.id]
        key = get_current_thread_key()
        del current_context[key]

    def state_save(self, state_directory: os.PathLike):
        path = Path(state_directory) / f"{self.id}.pickle"
        render_context = self.app_object
        if render_context is not None:
            render_context = cast(reacton.core._RenderContext, render_context)
            state = render_context.state_get()
            with path.open("wb") as f:
                logger.debug("State: %r", state)
                pickle.dump(state, f)


contexts: Dict[str, AppContext] = {}
# maps from thread key to AppContext, if AppContext is None, it exists, but is not set as current
current_context: Dict[str, Optional[AppContext]] = {}


def create_dummy_context():
    from . import kernel

    app_context = AppContext(
        id="dummy",
        kernel=kernel.Kernel(),
    )
    return app_context


def get_current_thread_key() -> str:
    thread = threading.currentThread()
    return get_thread_key(thread)


def get_thread_key(thread: threading.Thread) -> str:
    thread_key = thread._name + str(thread._ident)  # type: ignore
    return thread_key


def set_context_for_thread(context: AppContext, thread: threading.Thread):
    key = get_thread_key(thread)
    current_context[key] = context


def has_current_context() -> bool:
    thread_key = get_current_thread_key()
    return (thread_key in current_context) and (current_context[thread_key] is not None)


def get_current_context() -> AppContext:
    thread_key = get_current_thread_key()
    if thread_key not in current_context:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key}, but no known context found. This might be a bug in Solara. "
            f"(known contexts: {list(current_context.keys())}"
        )
    context = current_context[thread_key]
    if context is None:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key!r}, although the context is know, it was not set for this thread. "
            + "This might be a bug in Solara."
        )
    return context


def set_current_context(context: Optional[AppContext]):
    thread_key = get_current_thread_key()
    current_context[thread_key] = context


def _run_app(
    app_state,
    app_script: AppScript,
    pathname: str,
    render_context: reacton.core._RenderContext = None,
):

    # app.signal_hook_install()
    main_object = app_script.run()
    app_state = pickle.loads(app_state) if app_state is not None else None
    if app_state:
        logger.info("Restoring state: %r", app_state)

    context = get_current_context()
    container = context.container
    if isinstance(main_object, widgets.Widget):
        return main_object, render_context
    elif isinstance(main_object, Element) or isinstance(main_object, reacton.core.Component):
        if isinstance(main_object, Element):
            children = [main_object]
        else:
            children = [main_object()]
        solara_context = solara.RoutingProvider(children=children, routes=app_script.routes, pathname=pathname)
        if render_context is None:
            result = render(solara_context, container, handle_error=True, initial_state=app_state)
            # support older versions of react
            if isinstance(result, tuple):
                container, render_context = result
            else:
                render_context = result
        else:
            if app_state:
                render_context.state_set(render_context.context_root, app_state)
            result = render_context.render(solara_context)
            container = render_context.container
        # return container, render_context
    else:
        extra = ""
        dotted = []
        for key, value in vars(main_object).items():
            if isinstance(value, (Element, widgets.Widget)):
                dotted.append(f"{app_script.app_name}.{key}")
        if dotted:
            extra = " We did find that sub objects that might work: " + ", ".join(dotted)
        raise ValueError(
            f"Main object (with name {app_script.app_name} in {app_script.path}) is not a Widget, Element or Component, but {type(main_object)}." + extra
        )
    return container, render_context


def load_app_widget(app_state, app_script: AppScript, pathname: str):
    # load the app, and set it at the child of the context's container
    app_state_initial = app_state
    context = get_current_context()
    container = context.container
    assert container is not None
    try:
        render_context = context.app_object
        with context:
            app_state = app_state_initial
            with pdb_guard():
                widget, render_context = _run_app(
                    app_state,
                    app_script,
                    pathname,
                    render_context=render_context,
                )
                if render_context is None:
                    assert context.container is not None
                    context.container.children = [widget]

            if render_context:
                context.app_object = render_context

    except BaseException as e:
        error = ""
        error = "".join(traceback.format_exception(None, e, e.__traceback__))
        print(error, file=sys.stdout, flush=True)  # noqa
        # widget = widgets.Label(value="Error, see server logs")
        import html

        error = html.escape(error)
        with context:
            widget = widgets.HTML(f"<pre>{error}</pre>")
            container.children = [widget]


def solara_comm_target(comm, msg_first):
    app: Optional[AppScript] = None

    def on_msg(msg):
        nonlocal app
        data = msg["content"]["data"]
        method = data["method"]
        if method == "run":
            path = data.get("path", "")
            app_name = data.get("appName") or "__default__"
            app = apps[app_name]
            context = get_current_context()
            import ipyvuetify

            container = ipyvuetify.Html(tag="div")
            context.container = container
            with context:
                load_app_widget(None, app, path)
                comm.send({"method": "finished", "widget_id": context.container._model_id})
        elif method == "check":
            context = get_current_context()
        elif method == "reload":
            assert app is not None
            context = get_current_context()
            path = data.get("path", "")
            with context:
                load_app_widget(context.state, app, path)
                comm.send({"method": "finished"})

    comm.on_msg(on_msg)

    def reload():
        # we don't reload the app ourself, we send a message to the client
        # this ensures that we don't run code of any client that for some reason is connected
        # but not working anymore. And it indirectly passes a message from the current thread
        # (which is that of the Reloader/watchdog), to the thread of the client
        logger.debug(f"Send reload to client: {context.id}")
        comm.send({"method": "reload"})

    context = get_current_context()
    context.reload = reload


def register_solara_comm_target(kernel: Kernel):
    kernel.comm_manager.register_target("solara.control", solara_comm_target)


def initialize_virtual_kernel(context_id: str, websocket: websocket.WebsocketWrapper):
    kernel = Kernel()
    logger.info("new virtual kernel: %s", context_id)
    context = contexts[context_id] = AppContext(id=context_id, kernel=kernel, control_sockets=[], widgets={}, templates={})
    with context:
        widgets.register_comm_target(kernel)
        register_solara_comm_target(kernel)
        assert kernel is Kernel.instance()
        kernel.shell_stream = WebsocketStreamWrapper(websocket, "shell")
        kernel.control_stream = WebsocketStreamWrapper(websocket, "control")
        kernel.session.websockets.add(websocket)


from . import patch  # noqa

patch.patch()
# the default app (used in solara-server)
if "SOLARA_APP" in os.environ:
    with pdb_guard():
        apps["__default__"] = AppScript(os.environ.get("SOLARA_APP", "solara.website.pages:Page"))



"""
# use_route

See also [Understanding Routing](/docs/understanding/routing).


```python
def use_route() -> Tuple[Optional[solara.Route], List[solara.Route]]:
    ...
```

`use_route` returns (if found) the current route that matches the pathname, or None. It also returns all resolved routes of that level
(i.e. all siblings and itself). This return tuple is useful to build custom navigation (e.g. using tabs or buttons).


Routing starts with declaring a set of `routes` in your app (solara picks up the `routes` variable if it exists,
and it should be in the same namespace as `Page`).
In the demo below, we declared the following routes.

```python
routes = [
    solara.Route(path="/"),
    solara.Route(
        path="fruit",
        component=Fruit,
        children=[
            solara.Route(path="/"),
            solara.Route(path="kiwi"),
            solara.Route(path="banana"),
            solara.Route(path="apple"),
        ],
    ),
]
```

Note that all routes are relative, since a component does not know if it is embedded into a larger application, which may also do routing.
Therefore you should never use the `route.path` for navigation since the route object has no knowledge of the full url
(e.g. `/api/use_route/fruit/banana`) but only knows its small piece of the pathname (e.g. `banana`)

Use [`resolve_path`](/api/resolve_path) to request the full url for navigation, or simply use the `Link` component that can do this for us.

If the current route has children, any child component that calls `use_route` will return the matched route and its siblings of our children.


"""

import solara

title = "use_route"


@solara.component
def Fruit():
    # this gets all routes in fruit's children
    route, routes = solara.use_route()

    if route is None:
        with solara.Link("banana") as main:
            solara.Button("Fruit not found, go to banana")
        return main

    if route.path == "/":
        with solara.Link("banana") as main:
            solara.Button("Choose a fruit, I recomment banana")
        return main

    with solara.VBox() as main:
        with solara.HBox():
            for route_fruit in routes[1:]:
                with solara.Link(solara.resolve_path(route_fruit)):
                    solara.Button(route_fruit.path)

            with solara.Link("/api/use_route/fruit/nofruit", nofollow=True):
                solara.Button("Wrong fruit")
            with solara.Link("/api/use_route/not-routed", nofollow=True):
                solara.Button("Wrong url")
        solara.Success(f"You chose {route.path}")
    return main


@solara.component
def Page():
    # this gets the top level routes, '/' and 'fruit'
    route_current, routes_all = solara.use_route()
    with solara.VBox() as main:
        with solara.Card("Navigation using buttons"):
            with solara.HBox():
                for route in routes_all:
                    with solara.Link(route):
                        solara.Button(route.path, color="red" if route_current == route else None)
        with solara.Card("Content decided by route:"):
            if route_current is None:
                solara.Error("Page does not exist")
                with solara.Link("fruit/kiwi"):
                    solara.Button("Go to fruit/kiwi")
            elif route_current.path == "/":
                with solara.Link("fruit/banana"):
                    solara.Button("Go to fruit/banana")
            elif route_current.path == "fruit":
                Fruit()
            else:
                solara.Error(f"Unknown route: {route_current.path}")
    return main


routes = [
    solara.Route(path="/"),
    solara.Route(
        path="fruit",
        component=Fruit,
        children=[
            solara.Route(path="/"),
            solara.Route(path="kiwi"),
            solara.Route(path="banana"),
            solara.Route(path="apple"),
        ],
    ),
]

sources = [Fruit.f, Page.f]  # type: ignore



import json
import logging
import os
import platform
import threading
import time
import uuid
from typing import Dict, Optional
from urllib.parse import quote

import ipyvue
import ipyvuetify
import ipywidgets
import requests

import solara

from . import settings

logger = logging.getLogger("solara.server.telemetry")

_server_user_id_override = None
_server_start_time = time.time()
# Privacy note: mixpanel does not store the IP, only the region
_server_ip = None
_platform_system = platform.system()
_platform_release = platform.release()
_python_version = platform.python_version()

solara_props = {
    "solara_version": solara.__version__,
    "ipywidgets_version": ipywidgets.__version__,
    "ipyvuetify_version": ipyvuetify.__version__,
    "ipyvue_version": ipyvue.__version__,
}
_docker = False

try:
    path = "/proc/self/cgroup"
    _docker = os.path.exists("/.dockerenv") or os.path.isfile(path) and any("docker" in line for line in open(path))
except:  # noqa
    logger.exception("Failed to detect docker")


def _get_ip():
    global _server_ip
    try:
        _server_ip = requests.get("https://api.ipify.org").text
    except Exception:
        logger.exception("Failed to get server IP")
        _server_ip = "failed to get IP"


threading.Thread(target=_get_ip())


def override_server_user_id(server_user_id: str):
    global _server_user_id_override
    _server_user_id_override = server_user_id


def get_server_user_id():
    return _server_user_id_override or settings.telemetry.server_user_id


def track(event: str, props: Optional[Dict] = None):
    if not settings.telemetry.mixpanel_enable:
        return
    event_item = {
        "event": event,
        "properties": {
            "token": settings.telemetry.mixpanel_token,
            "fingerprint": settings.telemetry.server_fingerprint,
            "time": int(time.time() * 1000),
            "distinct_id": get_server_user_id(),
            # can be useful to get of session duration
            "session_id": settings.telemetry.server_session_id,
            "$insert_id": str(uuid.uuid4()),  # to de-duplicate events
            # Privacy note: mixpanel does not store the IP, only the region
            "ip": _server_ip,
            "platform_system": _platform_system,
            "platform_release": _platform_release,
            "python_version": _python_version,
            "docker": _docker,
            **(solara_props or {}),
            **(props or {}),
        },
    }
    try:
        requests.post(
            "https://api.mixpanel.com/track/",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data=f"data={quote(json.dumps([event_item]))}",
            timeout=1,
        )
    except Exception:
        logger.exception("Failed mixpanel API call")


def server_start():
    if _server_ip is None:
        # wait at most 0.1 second to get the server ip
        for i in range(10):
            time.sleep(0.01)
            if _server_ip is not None:
                break
        else:
            logger.warning("Server IP is not known yet")
    global _server_start_time
    _server_start_time = time.time()
    track("Solara server start")


def server_stop():
    duration = time.time() - _server_start_time
    track("Solara server stop", {"duration_seconds": duration})


if __name__ == "__main__":
    track("Solara test event", {"where": "command line"})



"""
# Select components

Select comes in two flavours:

   * `Select` for a singular selection
   * `SelectMultiple` which allows for multiple selections


"""
import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    with solara.VBox() as main:
        with solara.Card("Single selection"):
            food, set_food = solara.use_state("Banana")
            solara.Markdown(f"**Selected**: {food}")
            foods = ["Kiwi", "Banana", "Apple"]
            solara.Select(label="Food", value=food, values=foods, on_value=set_food)

        with solara.Card("Multiple selections"):
            all_languages = "Python C++ Java JavaScript TypeScript BASIC".split()
            languages, set_languages = solara.use_state([all_languages[0]])
            solara.Markdown(f"**Selected**: {languages}")
            solara.SelectMultiple("Languages", languages, all_languages, on_value=set_languages)

    return main


__doc__ += "# Select"
__doc__ += apidoc(solara.Select.f)  # type: ignore
__doc__ += "# SelectMultiple"
__doc__ += apidoc(solara.SelectMultiple.f)  # type: ignore



import json
import logging
from typing import Union

import js

from . import app, server
from .websocket import WebsocketWrapper

connection_id = "single"

logger = logging.getLogger("solara.server.pyodide")


class Websocket(WebsocketWrapper):
    def send_text(self, data: str) -> None:
        js.sendToPage(data)

    def send_bytes(self, data: bytes) -> None:
        js.sendToPage(data)

    def close(self) -> None:
        pass

    def receive(self) -> Union[str, bytes]:
        return b""


ws = Websocket()


def start(app_name):
    logger
    app.apps["__default__"].close()
    app.apps["__default__"] = app.AppScript(app_name)
    app.initialize_virtual_kernel(connection_id, ws)


async def processKernelMessage(msg):
    msg = json.loads(msg)
    context = app.contexts[connection_id]
    kernel = context.kernel
    with context:
        server.process_kernel_messages(kernel, msg)
    return 42



"""# Image
"""

from pathlib import Path

import numpy as np
import PIL.Image

import solara
import solara.website
from solara.website.utils import apidoc


@solara.component
def Page():

    image_path = Path(solara.website.__file__).parent / "public/beach.jpeg"
    image_url = "/static/public/beach.jpeg"
    image_ndarray = np.asarray(PIL.Image.open(image_path))

    with solara.VBox() as main:
        with solara.Card(title="As a path"):
            solara.Image(image_path)

        with solara.Card(title="As a URL"):
            solara.Image(image_url)

        with solara.Card(title="As NumPy array"):
            solara.Image(image_ndarray)

    return main


__doc__ += apidoc(solara.Image.f)  # type: ignore



"""Build webapps using IPywidgets"""
__version__ = "1.8.1"
github_url = "https://github.com/widgetti/solara"
git_branch = "master"


def _using_solara_server():
    import sys

    if "solara.server" in sys.modules:
        return True
    if sys.argv[0].split("/")[-1] == "solara":
        return True
    return False


# isort: skip_file
from reacton import (
    component,
    component_interactive,
    value_component,
    create_context,
    get_widget,
    make,
    provide_context,
    render,
    render_fixed,
    use_context,
    use_effect,
    use_exception,
    use_memo,
    use_reducer,
    use_ref,
    use_side_effect,
    use_state,
    use_state_widget,
)  # noqa: F403, F401
from reacton.core import Element  # noqa: F403, F401
from . import util

# flake8: noqa: F402
from .datatypes import *
from .hooks import *
from .cache import memoize
from . import cache
from .components import *

from .routing import use_route, use_router, use_route_level, find_route, use_pathname, resolve_path
from .autorouting import generate_routes, generate_routes_directory, RenderPage, RoutingProvider, DefaultLayout
from .checks import check_jupyter


def display(*objs, **kwargs):
    """Display a Python object in the current component.

    ## Supported objects

    Any object that can be displayed in the Jupyter notebook can be displayed in Solara
    as well. This means that plotly, pandas, vaex and other libraries that have
    a display function will work out of the box in Solara.

    However, if you require callback functions, use the specific Solara components, e.g.:

      * [Plotly](/api/plotly)
      * [Altair](/api/altair)
      * [Matplotlib](/api/matplotlib)
      * [Dataframe](/api/dataframe)

    ```solara
    import solara
    import pandas as pd
    import plotly.express as px

    df = px.data.gapminder()
    fig = px.scatter(df, x="gdpPercap", y="lifeExp")
    button = solara.Button("Click me")

    @solara.component
    def Page():

        # You can use solara.display directly
        solara.display(fig)

        # Or use the IPython display function, which is in the global scope
        # in Jupyter and Solara
        display(df)

        with solara.Card("Some card"):
            # you can use display to add existing elements to the parent's
            # children
            display(button)
    ```
    Note that this is a dispatch call to the
    [IPython display function](https://ipython.readthedocs.io/en/stable/api/generated/IPython.display.html#IPython.display.display),
    with a slimmer API. The main purpose of this function is the have type safety, since tools like mypy do now know about
    the existance of the IPython display function, and will complain if you use it without importing it.
    Since solara is always important, this saves you from having to import IPython in your code.

    ## Arguments

     * `objs`: The objects to display.
     * `kwargs`: Keyword arguments to pass to the IPython display function.

    """
    from IPython.display import display as ipy_display

    ipy_display(*objs, **kwargs)



from typing import Callable

import solara
from solara.alias import rv
from solara.server import settings


@solara.component
def Header(
    on_toggle_left_menu: Callable[[], None] = None,
    on_toggle_right_menu: Callable[[], None] = None,
):
    # use routes of parent (assuming we are a child of a layout)
    route_current, all_routes = solara.use_route(level=-1)
    router = solara.use_router()

    # set states for menu
    with rv.AppBar(tag="header", flat=True, class_="bg-primary-fade padding-40", height="auto") as main:
        with rv.ToolbarTitle(class_="d-flex", style_="align-items:center"):
            if route_current and len(route_current.children) > 0:
                with solara.Button(icon=True, class_="hidden-md-and-up", on_click=lambda: on_toggle_left_menu and on_toggle_left_menu()):
                    rv.Icon(children=["mdi-menu"])
            with solara.Link(path_or_route="/"):
                solara.Image(router.root_path + "/static/assets/images/logo.svg")
        rv.Spacer()

        if settings.search.enabled:
            from solara_enterprise.search.search import Search

            Search()

        # menu
        with rv.Html(tag="ul", class_="main-menu menu d-none d-md-flex"):
            for route in all_routes:
                if route.path == "apps":
                    continue
                current = route_current == route
                with rv.Html(tag="li", class_="active" if current else None):
                    solara.Link("/" + route.path if route.path != "/" else "/", children=[route.label])
        with rv.Btn(icon=True, tag="a", class_="d-none d-md-flex", attributes={"href": solara.github_url, "target": "_blank"}):
            rv.Icon(children=["mdi-github-circle"])

        with solara.Button(icon=True, class_="hidden-md-and-up", on_click=lambda: on_toggle_right_menu and on_toggle_right_menu()):
            rv.Icon(children=["mdi-menu"])

    return main



import contextlib
import json
import logging
import os
import sys
import time
from pathlib import Path
from typing import Dict, List, Optional, TypeVar

import ipykernel
import jinja2
import requests

import solara
import solara.routing

from . import app, settings, websocket
from .app import initialize_virtual_kernel
from .kernel import Kernel, deserialize_binary_message

COOKIE_KEY_SESSION_ID = "solara-session-id"

T = TypeVar("T")


directory = Path(__file__).parent
template_name = "index.html.j2"
ipykernel_major = int(ipykernel.__version__.split(".")[0])
cache_memory = solara.cache.Memory(max_items=128)

# first look at the project directory, then the builtin solara directory


def get_jinja_env(app_name: str) -> jinja2.Environment:
    jinja_loader = jinja2.FileSystemLoader(
        [
            app.apps["__default__"].directory.parent / "templates",
            str(directory / "templates"),
        ]
    )
    jinja_env = jinja2.Environment(loader=jinja_loader, autoescape=True)
    return jinja_env


logger = logging.getLogger("solara.server.server")
nbextensions_ignorelist = [
    "jupytext/index",
    "nbextensions_configurator/config_menu/main",
    "jupytext/index",
    "nbdime/index",
    "voila/extension",
    "contrib_nbextensions_help_item/main",
    "execute_time/ExecuteTime",
    "dominocode/extension",
    "low-code-assistant/extension",
    "domino-code-assist/extension",
    "jupyter-js/extension",
    "jupyter-js-widgets/extension",
    "jupyter_dash/main",
]


def readyz():
    return {"status": "ok"}, 200


def wait_ready(url, timeout=10) -> None:
    """Wait for a solara server at root url to be ready, or throw a TimeoutError

    This uses the /readyz endpoint to check if the server is ready.

    >>> solara.server.server.wait_ready("http://localhost:8888")
    ...

    """
    t0 = time.time()
    while True:
        try:
            r = requests.get(url + "/readyz")
            if r.status_code == 200:
                return
        except Exception:
            pass
        time.sleep(0.1)
        if time.time() - t0 > timeout:
            raise TimeoutError(f"Timeout waiting for {url}")


def is_ready(url) -> bool:
    """Returns wether a solara server at root url is ready.

    This uses the /readyz endpoint to check if the server is ready.

    >>> solara.server.server.is_ready("http://localhost:8888")
    True

    """
    try:
        r = requests.get(url + "/readyz")
        if r.status_code == 200:
            return True
    except Exception:
        pass
    return False


async def app_loop(ws: websocket.WebsocketWrapper, session_id: str, connection_id: str):
    initialize_virtual_kernel(connection_id, ws)
    context = app.contexts.get(connection_id)
    if context is None:
        logging.warning("invalid context id: %r", connection_id)
        # to avoid very fast reconnects (we are in a thread anyway)
        time.sleep(0.5)
        return

    if settings.main.tracer:
        import viztracer

        output_file = f"viztracer-{connection_id}.html"
        run_context = viztracer.VizTracer(output_file=output_file, max_stack_depth=10)
        logger.warning(f"Running with tracer: {output_file}")
    else:
        run_context = contextlib.nullcontext()

    kernel = context.kernel
    with run_context:
        while True:
            try:
                message = ws.receive()
            except websocket.WebSocketDisconnect:
                logger.debug("Disconnected")
                return
            t0 = time.time()
            if isinstance(message, str):
                msg = json.loads(message)
            else:
                msg = deserialize_binary_message(message)
            t1 = time.time()
            with context:
                process_kernel_messages(kernel, msg)
                t2 = time.time()
                if settings.main.timing:
                    print(f"timing: total={t2-t0:.3f}s, deserialize={t1-t0:.3f}s, kernel={t2-t1:.3f}s")  # noqa: T201


def process_kernel_messages(kernel: Kernel, msg: Dict):
    session = kernel.session
    comm_manager = kernel.comm_manager

    def send_status(status, parent):
        session.send(
            kernel.iopub_socket,
            "status",
            {"execution_state": status},
            parent=parent,
            ident=None,
        )

    @contextlib.contextmanager
    def busy_idle(parent):
        send_status("busy", parent=parent)
        try:
            yield
        finally:
            send_status("idle", parent=parent)

    msg_type = msg["header"]["msg_type"]

    if ipykernel_major < 6:
        # the channel argument was added in 6.0
        kernel.set_parent(None, msg["header"])
    else:
        kernel.set_parent(None, msg["header"], msg["channel"])
    if msg_type == "kernel_info_request":
        content = {
            "status": "ok",
            "protocol_version": "5.3",
            "implementation": Kernel.implementation,
            "implementation_version": Kernel.implementation_version,
            "language_info": {},
            "banner": Kernel.banner,
            "help_links": [],
        }
        with busy_idle(msg["header"]):
            msg = kernel.session.send(kernel.shell_stream, "kernel_info_reply", content, msg["header"], None)
    elif msg_type in ["comm_open", "comm_msg", "comm_close"]:
        with busy_idle(msg["header"]):
            getattr(comm_manager, msg_type)(kernel.shell_stream, None, msg)
    elif msg_type in ["comm_info_request"]:
        content = msg["content"]
        target_name = msg.get("target_name", None)

        comms = {k: dict(target_name=v.target_name) for (k, v) in comm_manager.comms.items() if v.target_name == target_name or target_name is None}
        reply_content = dict(comms=comms, status="ok")
        with busy_idle(msg["header"]):
            msg = session.send(
                kernel.shell_stream,
                "comm_info_reply",
                reply_content,
                msg["header"],
                None,
            )
    else:
        logger.error("Unsupported msg with msg_type %r", msg_type)


def read_root(path: str, root_path: str = "", render_kwargs={}, use_nbextensions=True) -> Optional[str]:
    default_app = app.apps["__default__"]
    routes = default_app.routes
    router = solara.routing.Router(path, routes)
    if not router.possible_match:
        return None
    if use_nbextensions:
        nbextensions = get_nbextensions()
    else:
        nbextensions = []

    resources = {
        "theme": "light",
        "nbextensions": nbextensions,
    }
    template: jinja2.Template = get_jinja_env(app_name="__default__").get_template(template_name)
    pre_rendered_html = ""
    pre_rendered_css = ""
    pre_rendered_metas = ""
    title = "Solara ☀️"
    if settings.ssg.enabled:
        from solara_enterprise import ssg

        ssg_data = ssg.ssg_data(path)
        if ssg_data is not None:
            pre_rendered_html = ssg_data["html"]
            pre_rendered_css = "\n".join(ssg_data["styles"])
            pre_rendered_metas = "\n    ".join(ssg_data["metas"])
            title = ssg_data["title"]

    if settings.assets.proxy:
        # solara acts as a proxy
        cdn = f"{root_path}/_solara/cdn"
    else:
        cdn = settings.assets.cdn

    render_settings = {
        "title": title,
        "path": path,
        "root_path": root_path,
        "resources": resources,
        "theme": settings.theme.dict(),
        "production": settings.main.mode == "production",
        "pre_rendered_html": pre_rendered_html,
        "pre_rendered_css": pre_rendered_css,
        "pre_rendered_metas": pre_rendered_metas,
        "assets": settings.assets.dict(),
        "cdn": cdn,
        **render_kwargs,
    }
    response = template.render(**render_settings)
    return response


def find_prefixed_directory(path):
    prefixes = [sys.prefix, os.path.expanduser("~/.local")]
    for prefix in prefixes:
        directory = f"{prefix}{path}"
        if Path(directory).exists():
            return directory
    else:
        raise RuntimeError(f"{path} not found at prefixes: {prefixes}")


@solara.memoize(storage=cache_memory)
def get_nbextensions_directories() -> List[Path]:
    from jupyter_core.paths import jupyter_path

    all_nb_directories = [Path(k) for k in jupyter_path("nbextensions")]
    # FIXME: remove IPython nbextensions path after a migration period
    try:
        from IPython.paths import get_ipython_dir
    except ImportError:
        pass
    else:
        all_nb_directories.append(Path(get_ipython_dir()) / "nbextensions")
    return [Path(k) for k in all_nb_directories]


@solara.memoize(storage=cache_memory)
def get_nbextensions() -> List[str]:
    from jupyter_core.paths import jupyter_config_path

    read_config_path = [os.path.join(p, "serverconfig") for p in jupyter_config_path()]
    read_config_path += [os.path.join(p, "nbconfig") for p in jupyter_config_path()]
    # import inline since we don't want this dep for pyiodide
    from jupyter_server.services.config import ConfigManager

    config_manager = ConfigManager(read_config_path=read_config_path)
    notebook_config = config_manager.get("notebook")
    # except for the widget extension itself, since Voilà has its own
    load_extensions = notebook_config.get("load_extensions", {})
    if "jupyter-js-widgets/extension" in load_extensions:
        load_extensions["jupyter-js-widgets/extension"] = False
    if "voila/extension" in load_extensions:
        load_extensions["voila/extension"] = False
    if "voila/extension" in load_extensions:
        load_extensions["voila/extension"] = False
    directories = get_nbextensions_directories()

    def exists(name):
        for directory in directories:
            if (directory / (name + ".js")).exists():
                return True
        logger.error(f"nbextension {name} not found")
        return False

    nbextensions = [name for name, enabled in load_extensions.items() if enabled and (name not in nbextensions_ignorelist) and exists(name)]
    return nbextensions


if "pyodide" not in sys.modules:
    nbextensions_directories = get_nbextensions_directories()
    nbconvert_static = find_prefixed_directory("/share/jupyter/nbconvert/templates/lab/static")
    solara_static = Path(__file__).parent / "static"



# tests all pages in the api docs
from pathlib import Path

import playwright.sync_api

import solara

HERE = Path(__file__).parent


md = """

# Solara markdown

```solara
import solara


@solara.component
def ClickButton():
    clicks, set_clicks = solara.use_state(0)
    def on_click():
        set_clicks(clicks + 1)
        print("clicks", clicks)

    return solara.Button(label=f"Clicked: {clicks}", on_click=on_click)


app = ClickButton()
```

"""


@solara.component
def MarkdownApp():
    return solara.Markdown(md, unsafe_solara_execute=True)


@solara.component
def MarkdownAppOff():
    return solara.Markdown(md, unsafe_solara_execute=False)


@solara.component
def MarkdownItApp():
    return solara.MarkdownIt(md, unsafe_solara_execute=True)


@solara.component
def MarkdownItAppOff():
    return solara.MarkdownIt(md, unsafe_solara_execute=False)


def test_markdown(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("markdown_test:MarkdownApp"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Clicked: 0").wait_for()
        # TODO: flakey test
        page_session.locator("text=Clicked: 0").click()
        # page_session.locator("text=Clicked: 1").wait_for()


def test_markdown_no_execute(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("markdown_test:MarkdownAppOff"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Solara execution is not enabled").wait_for()


def test_markdown_it(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("markdown_test:MarkdownItApp"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Clicked: 0").wait_for()
        page_session.locator("text=Clicked: 0").click()
        # TODO: flakey test
        # page_session.locator("text=Clicked: 1").wait_for()


def test_markdown_it_no_execute(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("markdown_test:MarkdownItAppOff"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Solara execution is not enabled").wait_for()



"""# Input fields

This page contains all the input fields available in Solara.

# InputText
"""
import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    text, set_text = solara.use_state("Hello world")
    password, set_password = solara.use_state("Super secret")
    number, set_number = solara.use_state(42.0)

    continuous_update, set_continuous_update = solara.use_state(False)

    with solara.VBox() as main:
        with solara.Card("Settings"):
            solara.Checkbox(label="Continuous update", value=continuous_update, on_value=set_continuous_update)

        with solara.Card("Text"):
            solara.InputText("Enter some text", value=text, on_value=set_text, continuous_update=continuous_update)
            with solara.HBox():
                solara.Button("Clear", on_click=lambda: set_text(""))
                solara.Button("Reset", on_click=lambda: set_text("Hello world"))
            solara.Markdown(f"**You entered**: {text}")

        with solara.Card("Password"):
            solara.InputText("Enter a password", value=password, on_value=set_password, password=True, continuous_update=continuous_update)
            with solara.HBox():
                solara.Button("Clear", on_click=lambda: set_password(""))
                solara.Button("Reset", on_click=lambda: set_password("Super secret"))
            solara.Markdown(f"**You entered**: {password}")

        with solara.Card("Number (float)"):
            solara.InputFloat("Enter some number", value=number, on_value=set_number, continuous_update=continuous_update)
            with solara.HBox():
                solara.Button("Reset", on_click=lambda: set_number(42))
            solara.Markdown(f"**You entered**: {number}")

    return main


__doc__ += apidoc(solara.InputText.f)  # type: ignore

__doc__ += "# InputFloat"
__doc__ += apidoc(solara.InputFloat.f)  # type: ignore



from typing import Callable, List

from reacton import ipyvue
from reacton import ipyvuetify as v

import solara
import solara.util


@solara.component
def Button(
    label: str = None,
    on_click: Callable[[], None] = None,
    icon_name: str = None,
    children: list = [],
    disabled=False,
    text=False,
    click_event="click",
    classes: List[str] = [],
    value=None,
    **kwargs,
):
    """A button that can be clicked to trigger an event.


    ## Arguments

    - `label`: The text to display on the button.
    - `on_click`: A callback function that is called when the button is clicked.
    - `icon_name`: The name of the icon to display on the button ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
    - `children`: A list of child elements to display on the button.
    - `disabled`: Whether the button is disabled.
    - `value`: (Optional) When used as a child of a ToggleButtons component, the value of the selected button, see [ToggleButtons](/api/togglebuttons).
    - `classes`: additional CSS classes to apply.

    ### Deprecated arguments
    - click_event: (Deprecated/export option: The event that triggers the on_click callback, which can include vue event modifiers).

    """
    if label:
        children = [label] + children
    if icon_name:
        children = [v.Icon(left=bool(label), children=[icon_name])] + children
    if "class_" in kwargs:
        kwargs = kwargs.copy()
        class_ = solara.util._combine_classes([*classes, kwargs.pop("class_")])
    else:
        class_ = solara.util._combine_classes(classes)
    btn = v.Btn(children=children, **kwargs, disabled=disabled, text=text, class_=class_)
    ipyvue.use_event(btn, click_event, lambda *_ignore: on_click and on_click())
    return btn



import solara

github_url = solara.util.github_url(__file__)


@solara.component
def Page():
    solara.Title("Page 2 set with title")
    with solara.Card("Page 2"):
        solara.Markdown("Page 2 is even better, even though it has no sidebar.")
        solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)



"""# generate_routes

"""
import solara
import solara.autorouting
from solara.website.utils import apidoc

from . import NoPage

title = "generate_routes"
Page = NoPage
__doc__ += apidoc(solara.autorouting.generate_routes)  # type: ignore



import linecache
from pathlib import Path
from typing import Any, Dict

import nbformat

import solara
import solara.components.applayout

HERE = Path(__file__).parent


def execute_notebook(path: Path):
    nb: nbformat.NotebookNode = nbformat.read(path, 4)
    stat = path.stat()
    scope: Dict[str, Any] = {}
    for cell_index, cell in enumerate(nb.cells):
        cell_index += 1  # used 1 based
        cell.scope_snapshot = {}
        if cell.cell_type == "code":
            if cell.source.startswith("## solara: skip"):
                continue
            source = cell.source
            cell_path = f"app.ipynb input cell {cell_index}"
            ast = compile(source, cell_path, "exec")
            entry = (
                stat.st_size,
                stat.st_mtime,
                [line + "\n" for line in source.splitlines()],
                cell_path,
            )
            linecache.cache[cell_path] = entry
            exec(ast, scope)
            # every cell gets a snapshot of the scope
            scope_snapshot = {**scope}
            # avoid the buggy setitem of NotebookNode
            # which causes an recursion error
            cell.__dict__["scope_snapshot"] = scope_snapshot
        elif cell.cell_type == "markdown":
            pass
        else:
            raise ValueError(f"Unknown cell type: {cell.cell_type}, supported types are: code, markdown")
    return nb


@solara.component
def Page():
    # only execute once, other
    nb: nbformat.NotebookNode = solara.use_memo(lambda: execute_notebook(HERE / "_data_science.ipynb"))

    last_page = None

    solara.components.applayout.should_use_embed.provide(True)
    with solara.VBox() as main:
        for cell_index, cell in list(enumerate(nb.cells)):
            cell_index += 1  # used 1 based
            if cell.cell_type == "code":
                if cell.source.startswith("## solara: skip"):
                    continue
                scope = cell.scope_snapshot
                page = scope.get("Page")
                solara.Markdown(
                    f"""
```python
{cell.source}
```"""
                )
                if page != last_page and page is not None:
                    with solara.AppLayout(navigation=False, toolbar_dark=False):
                        page()
                last_page = page
            elif cell.cell_type == "markdown":
                solara.Markdown(cell.source)
            else:
                raise ValueError(f"Unknown cell type: {cell.cell_type}, supported types are: code, markdown")
    return main



import asyncio
import logging
import mimetypes
import os
from http.server import HTTPServer
from pathlib import Path
from typing import Any
from uuid import uuid4

import flask
import simple_websocket
import solara
from flask import Blueprint, Flask, request, send_from_directory, url_for
from flask_sock import Sock
from solara.server.threaded import ServerBase

from . import app as appmod
from . import cdn_helper, server, settings, websocket

os.environ["SERVER_SOFTWARE"] = "solara/" + str(solara.__version__)

logger = logging.getLogger("solara.server.flask")
blueprint = Blueprint("blueprint-solara", __name__)
websocket_extension = Sock()


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: simple_websocket.Server

    def __init__(self, ws: simple_websocket.Server) -> None:
        self.ws = ws

    def close(self):
        self.ws.close()

    def send_text(self, data: str) -> None:
        self.ws.send(data)

    def send_bytes(self, data: bytes) -> None:
        self.ws.send(data)

    def receive(self):
        return self.ws.receive()


class ServerFlask(ServerBase):
    server: Any

    def has_started(self):
        return server.is_ready(f"http://{self.host}:{self.port}")

    def signal_stop(self):
        assert isinstance(self.server, HTTPServer)
        self.server.shutdown()  # type: ignore

    def serve(self):
        from solara.server.flask import app
        from werkzeug.serving import make_server

        self.server = make_server(self.host, self.port, app, threaded=True)  # type: ignore
        assert isinstance(self.server, HTTPServer)
        self.started.set()
        self.server.serve_forever(poll_interval=0.05)  # type: ignore


@blueprint.route("/jupyter/api/kernels/<id>")
def kernels(id):
    return {"name": "lala", "id": "dsa"}


@websocket_extension.route("/jupyter/api/kernels/<id>/<name>")
def kernels_connection(ws: simple_websocket.Server, id: str, name: str):
    try:
        connection_id = request.args["session_id"]
        session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID)
        logger.info("Solara kernel requested for session_id=%s connection_id=%s", session_id, connection_id)
        if session_id is None:
            logger.error("no session cookie")
            ws.close()
            return
        ws_wrapper = WebsocketWrapper(ws)
        asyncio.run(server.app_loop(ws_wrapper, session_id=session_id, connection_id=connection_id))
    except simple_websocket.ws.ConnectionClosed:
        pass  # ok
    except:  # noqa
        logger.exception("Error in kernel handler")
        raise


@blueprint.route("/_solara/api/close/<connection_id>", methods=["GET", "POST"])
def close(connection_id: str):
    if connection_id in appmod.contexts:
        context = appmod.contexts[connection_id]
        context.close()
    return ""


@blueprint.route("/static/public/<path:path>")
def public(path):
    directories = [app.directory.parent / "public" for app in appmod.apps.values()]
    for directory in directories:
        file = directory / path
        if file.exists():
            return send_from_directory(directory, path)
    return flask.Response("not found", status=404)


@blueprint.route("/static/assets/<path:path>")
def assets(path):
    overrides = [app.directory.parent / "assets" for app in appmod.apps.values()]
    default = server.solara_static.parent / "assets"
    directories = [*overrides, default]
    for directory in directories:
        file = directory / path
        if file.exists():
            return send_from_directory(directory, path)
    return flask.Response("not found", status=404)


@blueprint.route("/static/nbextensions/<dir>/<filename>")
def nbext(dir, filename):
    for directory in server.nbextensions_directories:
        file = directory / dir / filename
        if file.exists():
            return send_from_directory(directory / dir, filename)
    return flask.Response("not found", status=404)


@blueprint.route("/static/nbconvert/<path:path>")
def serve_nbconvert_static(path):
    return send_from_directory(server.nbconvert_static, path)


@blueprint.route("/static/<path:path>")
def serve_static(path):
    return send_from_directory(server.solara_static, path)


@blueprint.route(f"/{cdn_helper.cdn_url_path}/<path:path>")
def cdn(path):
    cache_directory = settings.assets.proxy_cache_dir
    content = cdn_helper.get_data(Path(cache_directory), path)
    mime = mimetypes.guess_type(path)
    return flask.Response(content, mimetype=mime[0])


@blueprint.route("/", defaults={"path": ""})
@blueprint.route("/<path:path>")
def read_root(path):
    root_path = url_for(".read_root")
    if root_path.endswith("/"):
        root_path = root_path[:-1]
    session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID) or str(uuid4())
    content = server.read_root(flask.request.path, root_path=root_path)
    if content is None:
        return flask.Response("not found", status=404)
    assert session_id is not None
    response = flask.Response(content, mimetype="text/html")
    response.set_cookie(server.COOKIE_KEY_SESSION_ID, value=session_id)
    return response


@blueprint.route("/readyz")
def readyz():
    json, status = server.readyz()
    return flask.Response(json, mimetype="application/json", status=status)


# using the blueprint and websocket blueprint makes it easier to integrate into other applications
websocket_extension.init_app(blueprint)
app = Flask(__name__)
app.register_blueprint(blueprint)

if __name__ == "__main__":
    app.run(debug=False, port=8765)



import contextlib
import logging

# import subprocess
from pathlib import Path

# import playwright.sync_api

# import solara.server.server

# from . import conftest

app_path = Path(__file__).parent / "testapp.py"

logger = logging.getLogger("solara-test.integration.reload_test")


HERE = Path(__file__).parent


@contextlib.contextmanager
def append(text):
    with app_path.open() as f:
        content = f.read()
    try:
        with app_path.open("w") as f:
            f.write(content)
            f.write(text)
        yield
    finally:
        with app_path.open("w") as f:
            f.write(content)


@contextlib.contextmanager
def replace(path, text):
    with path.open() as f:
        content = f.read()
    try:
        with path.open("w") as f:
            f.write(text)
        yield
    finally:
        with path.open("w") as f:
            f.write(content)


# button_code = """
# @solara.component
# def ButtonClick():
#     clicks, set_clicks = solara.use_state(Clicks(0))
#     return rw.Button(description=f"!!! {clicks.value} times", on_click=lambda: set_clicks(Clicks(clicks.value + 1)))


# app = ButtonClick()
# """


# def test_reload_with_pickle(page_session: playwright.sync_api.Page):
#     port = conftest.TEST_PORT + 1000
#     conftest.TEST_PORT += 1
#     args = ["solara", "run", f"--port={port}", "--no-open", "tests.integration.testapp:app", "--log-level=debug"]
#     popen = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
#     host = "localhost"
#     try:
#         solara.server.server.wait_ready(f"http://{host}:{port}", timeout=10)
#         page_session.goto(f"http://localhost:{port}")
#         page_session.locator("text=Clicked 0 times").click(timeout=5000)
#         page_session.locator("text=Clicked 1 times").click(timeout=5000)
#         page_session.locator("text=Clicked 2 times").wait_for(timeout=5000)
#         page_session.wait_for_timeout(1000)
#         with append(button_code):
#             page_session.locator("text=!!! 0 times").click(timeout=5000)
#             page_session.locator("text=!!! 1 times").click(timeout=5000)
#         popen.kill()
#     except Exception as e:
#         try:
#             popen.kill()
#         except:  # noqa
#             pass
#         outs, errs = popen.communicate(timeout=5)
#         if errs:
#             print("STDERR:")  # noqa
#             print(errs.decode("utf-8"))  # noqa
#         if outs:
#             print("STDOUT:")  # noqa
#             print(outs.decode("utf-8"))  # noqa
#         if errs:
#             raise ValueError("Expected no errors in solara server output") from e
#         raise


# the following tests are flakey on CI

# def test_reload_syntax_error(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(app_path.parent), solara_app("testapp:ButtonClick"):
#         # use as module, otherwise pickle wil not work
#         page_session.goto(solara_server.base_url)
#         assert page_session.title() == "Solara ☀️"
#         page_session.locator("text=Clicked 0 times").click()
#         page_session.locator("text=Clicked 1 times").click()

#         with append("\n$%#$%"):
#             reload.reloader.reload_event_next.wait()
#             # page_session.locator("text=Clicked 2 times").click()
#             page_session.locator("text=SyntaxError").wait_for()
#         reload.reloader.reload_event_next.wait()
#         page_session.locator("text=Clicked 2 times").click()
#         page_session.locator("text=Clicked 3 times").wait_for()


# def test_reload_change(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(app_path.parent), solara_app("testapp:app"):
#         logger.info("test_reload_many:run app")
#         page_session.goto(solara_server.base_url)
#         # assert page_session.title() == "Solara ☀️"
#         page_session.locator("text=Clicked 0 times").click()
#         page_session.locator("text=Clicked 1 times").click()
#         with append(button_code):
#             reload.reloader.reload_event_next.wait()
#             # page_session.locator("text=Clicked 2 times").click()
#             # page_session.locator("text=SyntaxError").wait_for()
#             page_session.locator("text=!!! 0 times").click()


# def test_reload_many(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(app_path.parent), solara_app("testapp:app"):
#         logger.info("test_reload_many:run app")
#         # use as module, otherwise pickle wil not work
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=Clicked 0 times").click()
#         page_session.locator("text=Clicked 1 times").click()

#         logger.info("test_reload_many:Touch app 1st time")
#         app_path.touch()
#         reload.reloader.reload_event_next.wait()
#         page_session.locator("text=Clicked 2 times").click()
#         page_session.locator("text=Clicked 3 times").wait_for(state="visible")

#         logger.info("test_reload_many:Touch app 2st time")
#         app_path.touch()
#         reload.reloader.reload_event_next.wait()
#         page_session.locator("text=Clicked 3 times").click()
#         page_session.locator("text=Clicked 4 times").wait_for(state="visible")


# def test_reload_vue(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(app_path.parent), solara_app("testapp:VueTestApp"):
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=foobar").wait_for()

#         vuecode = """
# <template>
#   <div>RELOADED</div>
# </template>
#         """
#         vuepath = Path(__file__).parent / "test.vue"
#         with replace(vuepath, vuecode):
#             page_session.locator("text=RELOADED").wait_for()
#         page_session.locator("text=foobar").wait_for()



"""# Interactive sine wave


This example shows how to have two slider control a visualization.

"""

import numpy as np
import plotly.express as px

import solara

x = np.linspace(0, 2, 100)

title = "Interactive sine wave"


@solara.component
def Page():
    freq, set_freq = solara.use_state(2.0)
    phase, set_phase = solara.use_state(0.1)
    y = np.sin(x * freq + phase)

    with solara.VBox() as main:
        solara.FloatSlider("Frequency", value=freq, on_value=set_freq, min=0, max=10)
        solara.FloatSlider("Phase", value=phase, on_value=set_phase, min=0, max=np.pi, step=0.1)

        fig = px.line(x=x, y=y)
        solara.FigurePlotly(fig)
    return main



import dataclasses
import unittest.mock
from typing import Callable, Dict, List, Optional, TypeVar

import ipyvuetify as v
import react_ipywidgets as react
from typing_extensions import TypedDict

import solara
import solara as sol
from solara.lab import State
from solara.lab.toestand import Reactive, Ref, use_sync_external_store
from solara.server import app, kernel

from ..common import click


@dataclasses.dataclass(frozen=True)
class Bears:
    type: str
    count: int = dataclasses.field()


B = TypeVar("B", bound=Bears)


class BearReactive(Reactive[B]):
    def increase_population(self):
        self.update(count=self.get().count + 1)


bears: Bears = Bears(type="brown", count=1)


def test_store_bare():
    # no need for subclasses
    mock = unittest.mock.Mock()
    mock_change = unittest.mock.Mock()
    store = Reactive[dict]({"string": "foo", "int": 42})
    unsub = store.subscribe(mock)
    unsub_change = store.subscribe_change(mock_change)
    mock.assert_not_called()
    mock_change.assert_not_called()
    store.update(string="bar")
    mock.assert_called_with({"string": "bar", "int": 42})
    mock_change.assert_called_with({"string": "bar", "int": 42}, {"string": "foo", "int": 42})

    unsub()
    unsub_change()


def test_subscribe():
    bear_store = BearReactive(bears)
    mock = unittest.mock.Mock()
    mock_type = unittest.mock.Mock()
    mock_count = unittest.mock.Mock()
    unsub = []
    unsub += [bear_store.subscribe(mock)]
    unsub += [Ref(bear_store.fields.type).subscribe(mock_type)]
    unsub += [Ref(bear_store.fields.count).subscribe(mock_count)]
    mock.assert_not_called()
    bear_store.update(type="purple")
    mock.assert_called_with(Bears(type="purple", count=1))
    mock_type.assert_called_with("purple")
    mock_count.assert_not_called()

    mock_type.reset_mock()
    bear_store.update(count=3)
    mock.assert_called_with(Bears(type="purple", count=3))
    mock_type.assert_not_called()
    mock_count.assert_called_with(3)
    for u in unsub:
        u()


def test_nested_update():
    # this effectively test the RLock vs Lock
    bear_store = BearReactive(bears)

    mock = unittest.mock.Mock()
    mock_type = unittest.mock.Mock()
    mock_count = unittest.mock.Mock()
    unsub = []
    unsub += [bear_store.subscribe(mock)]
    unsub += [Ref(bear_store.fields.type).subscribe(mock_type)]
    unsub += [Ref(bear_store.fields.count).subscribe(mock_count)]

    def reset_count(new_type):
        bear_store.update(count=0)

    Ref(bear_store.fields.type).subscribe(reset_count)
    Ref(bear_store.fields.type).value = "purple"
    mock.assert_called_with(Bears(type="purple", count=0))
    mock_type.assert_called_with("purple")
    mock_count.assert_called_with(0)
    for u in unsub:
        u()


def test_store_nested():
    @dataclasses.dataclass(frozen=True)
    class City:
        name: str
        population: int

    @dataclasses.dataclass(frozen=True)
    class Person:
        name: str
        height: float

    @dataclasses.dataclass(frozen=True)
    class Country:
        dicator: Person
        name: str
        cities: List[City] = dataclasses.field(default_factory=list)
        people: Dict[str, Person] = dataclasses.field(default_factory=dict)

    # no need for subclasses
    mock = unittest.mock.Mock()
    people = {"Jos": Person(name="Jos", height=1.8)}
    people_copy = people.copy()
    nl = Country(
        name="Netherlands",
        cities=[City(name="Amsterdam", population=1000000)],
        people=people,
        dicator=Person(name="Jos", height=1.8),
    )
    store = Reactive[Country](nl)
    unsub = store.subscribe(mock)
    mock.assert_not_called()
    country = store.fields
    population_accessor = Ref(store.fields.cities[0].population)
    assert population_accessor.get() == 1000000
    population_accessor.set(10)
    assert population_accessor.get() == 10

    mock.assert_called_with(
        Country(
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people=people_copy,
            dicator=Person(name="Jos", height=1.8),
        )
    )

    Ref(country.people["Jos"].height).set(1.9)
    assert people == people_copy
    mock.assert_called_with(
        Country(
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={"Jos": Person(name="Jos", height=1.9)},
            dicator=Person(name="Jos", height=1.8),
        )
    )

    Ref(country.people["Jos"].height).value = 2.1
    assert people == people_copy
    mock.assert_called_with(
        Country(
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={"Jos": Person(name="Jos", height=2.1)},
            dicator=Person(name="Jos", height=1.8),
        )
    )

    # TODO: new name?
    # Ref(country.people["Jos"].height).set(lambda x: x + 0.1)
    # mock.assert_called_with(
    #     Country(
    #         name="Netherlands",
    #         cities=[City(name="Amsterdam", population=10)],
    #         people={"Jos": Person(name="Jos", height=2.0)},
    #         dicator=Person(name="Jos", height=1.8),
    #     ),
    # )

    Ref(country.people).update(lambda x: {**x, "Maria": Person(name="Maria", height=1.7)})
    mock.assert_called_with(
        Country(
            dicator=Person(name="Jos", height=1.8),
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={
                "Jos": Person(name="Jos", height=2.1),
                "Maria": Person(name="Maria", height=1.7),
            },
        )
    )
    Ref(country.dicator.height).set(2.0)
    mock.assert_called_with(
        Country(
            dicator=Person(name="Jos", height=2.0),
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={
                "Jos": Person(name="Jos", height=2.1),
                "Maria": Person(name="Maria", height=1.7),
            },
        )
    )

    unsub()


def test_bear_store_basics():
    mock = unittest.mock.Mock()
    bear_store = BearReactive(bears)
    unsub = bear_store.subscribe(mock)
    mock.assert_not_called()
    bear_store.increase_population()
    mock.assert_called_with(Bears(type="brown", count=2))
    bear_store.increase_population()
    assert mock.call_count == 2
    mock.assert_called_with(Bears(type="brown", count=3))

    setter = bear_store.setter(bear_store.fields.count)
    setter(5)
    assert mock.call_count == 3
    mock.assert_called_with(Bears(type="brown", count=5))

    unsub()
    bear_store.increase_population()
    assert mock.call_count == 3

    # now test a subfield
    mock_count = unittest.mock.Mock()
    count = Ref(bear_store.fields.count)
    bear_store.subscribe(mock)
    count.subscribe(mock_count)
    count.set(4)
    assert count.get() == 4
    assert bear_store.get() == Bears(type="brown", count=4)
    mock.assert_called_with(Bears(type="brown", count=4))
    mock_count.assert_called_with(4)


def test_bear_store_basics_dict():
    class Bears(TypedDict):
        type: str
        count: int

    bears = Bears(type="brown", count=1)

    class BearReactive(Reactive[Bears]):
        def increase_population(self):
            self.update(count=self.get()["count"] + 1)

    mock = unittest.mock.Mock()
    bear_store = BearReactive(bears)
    unsub = bear_store.subscribe(mock)
    mock.assert_not_called()
    bear_store.increase_population()
    mock.assert_called_with(Bears(type="brown", count=2))
    bear_store.increase_population()
    assert mock.call_count == 2
    mock.assert_called_with(Bears(type="brown", count=3))

    setter = bear_store.setter(bear_store.fields["count"])
    setter(5)
    assert mock.call_count == 3
    mock.assert_called_with(Bears(type="brown", count=5))

    unsub()
    bear_store.increase_population()
    assert mock.call_count == 3


def test_bear_store_react():
    bear_store = BearReactive(bears)

    @react.component
    def BearCounter():
        bear_count = bear_store.use(lambda bear: bear.count)
        return sol.Info(f"{bear_count} bears around here")

    @react.component
    def Controls():
        return sol.Button("add bear", on_click=bear_store.increase_population)

    @react.component
    def App():
        with sol.VBox() as main:
            BearCounter()
            Controls()
        return main

    el = App()
    box, rc = react.render(el, handle_error=False)
    assert rc._find(v.Alert).widget.children[0] == "1 bears around here"
    click(rc._find(v.Btn).widget)
    assert rc._find(v.Alert).widget.children[0] == "2 bears around here"

    # the storage should live in the app context to support multiple users/connections
    kernel_shared = kernel.Kernel()
    context1 = app.AppContext(id="1", kernel=kernel_shared, control_sockets=[], widgets={}, templates={})
    context2 = app.AppContext(id="2", kernel=kernel_shared, control_sockets=[], widgets={}, templates={})
    for context in [context1, context2]:
        with context:
            rc.render(el)
            # el = App()
            # box, rc = react.render(el, handle_error=False)
            assert rc._find(v.Alert).widget.children[0] == "1 bears around here"
            click(rc._find(v.Btn).widget)
            assert rc._find(v.Alert).widget.children[0] == "2 bears around here"

    with context2:
        rc.render(el)
        assert rc._find(v.Alert).widget.children[0] == "2 bears around here"
        click(rc._find(v.Btn).widget)
        assert rc._find(v.Alert).widget.children[0] == "3 bears around here"

    with context1:
        rc.render(el)
        assert rc._find(v.Alert).widget.children[0] == "2 bears around here"
        click(rc._find(v.Btn).widget)
        assert rc._find(v.Alert).widget.children[0] == "3 bears around here"
        click(rc._find(v.Btn).widget)
        assert rc._find(v.Alert).widget.children[0] == "4 bears around here"
        click(rc._find(v.Btn).widget)
        assert rc._find(v.Alert).widget.children[0] == "5 bears around here"

    with context2:
        rc.render(el)
        assert rc._find(v.Alert).widget.children[0] == "3 bears around here"
        click(rc._find(v.Btn).widget)
        assert rc._find(v.Alert).widget.children[0] == "4 bears around here"


def test_simplest():

    settings = State({"bears": 2, "theme": "dark"})
    unsub = settings.subscribe(print)  # noqa

    settings.update(theme="light")
    # prints: {"bears": 2, "theme": "light"}

    unsub()  # remove event listener
    theme_accessor = Ref(settings.fields["theme"])

    # Now use it in a React component

    import react_ipywidgets as react

    renders = 0

    import solara as sol

    @react.component
    def ThemeInfo():
        # the lambda function here is called a selector, it 'selects' out the state you want
        # theme = settings.use(lambda state: state["theme"])
        theme = Ref(settings.fields["theme"]).use_value()
        return sol.Info(f"Using theme {theme}")

    @react.component
    def ThemeSelector():
        theme, set_theme = Ref(settings.fields["theme"]).use_state()
        with sol.ToggleButtonsSingle(theme, on_value=set_theme) as main:
            sol.Button("dark")
            sol.Button("light")
        return main

    @react.component
    def Test():
        nonlocal renders
        renders += 1
        with sol.VBox() as main:
            ThemeInfo()
            ThemeSelector()
        return main

    box, rc = react.render(Test(), handle_error=False)
    assert rc._find(v.Alert).widget.children[0] == "Using theme light"
    assert rc._find(v.BtnToggle).widget.v_model == 1
    theme_accessor.set("dark")
    assert rc._find(v.Alert).widget.children[0] == "Using theme dark"
    assert rc._find(v.BtnToggle).widget.v_model == 0
    rc._find(v.BtnToggle).widget.v_model = 1
    assert rc._find(v.Alert).widget.children[0] == "Using theme light"
    renders_before = renders
    settings.update(bears=3)
    assert renders == renders_before


@dataclasses.dataclass(frozen=True)
class Fish:
    fishes: int


F = TypeVar("F", bound=Fish)


class FishReactive(Reactive[F]):
    def jump(self):
        print("jump")  # noqa


@dataclasses.dataclass(frozen=True)
class AppStateComposite:
    bear: Bears
    fish: Fish


class AppState(Reactive[AppStateComposite]):
    bears: BearReactive

    def __post__init__(self):
        # for composite stores, manually create the substores
        self.bears = BearReactive(Ref(self.fields.bear))
        # it's ok not to have a substore for fish

    def eat(self):
        state = self.get()
        bears = state.bear.count
        fish = Fish(fishes=max(0, state.fish.fishes - bears))
        self.update(fish=fish)


def test_store_primitive():
    string_value = State("foo")
    mock = unittest.mock.Mock()
    unsub = string_value.subscribe(mock)
    mock.assert_not_called()
    string_value.set("bar")
    mock.assert_called_with("bar")
    unsub()


def test_store_computed():
    list_store = Reactive[list]([1, 2, 3])

    count = list_store.computed(len)
    last = list_store.computed(lambda x: x[-1] if x else None)

    assert count.get() == 3
    assert last.get() == 3
    mock = unittest.mock.Mock()
    mock_last = unittest.mock.Mock()
    unsub = count.subscribe(mock)
    unsub_last = last.subscribe(mock_last)
    mock.assert_not_called()
    mock_last.assert_not_called()
    list_store.set([1, 2, 3, 42])
    mock.assert_called_with(4)
    mock_last.assert_called_with(42)
    unsub()
    unsub_last()
    list_store.set([])
    mock.assert_called_with(4)
    mock_last.assert_called_with(42)


def test_app_composite():
    mock = unittest.mock.Mock()
    fish_state = Fish(fishes=4)
    app_state = AppStateComposite(bear=bears, fish=fish_state)
    app_store = AppState(app_state)
    app_store.subscribe(mock)

    assert app_store.get().bear.count == 1
    assert app_store.bears.get().count == 1
    assert app_store.get().fish.fishes == 4

    mock.assert_not_called()
    app_store.eat()
    assert app_store.get().fish.fishes == 3
    assert mock.call_count == 1
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=1), fish=Fish(fishes=3)))

    app_store.bears.increase_population()
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=2), fish=Fish(fishes=3)))
    assert mock.call_count == 2
    assert app_store.get().bear.count == 2
    assert app_store.bears.get().count == 2
    assert app_store.get().fish.fishes == 3

    app_store.eat()
    assert app_store.get().fish.fishes == 1
    assert mock.call_count == 3
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=2), fish=Fish(fishes=1)))
    app_store.eat()
    assert app_store.get().fish.fishes == 0
    assert mock.call_count == 4
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=2), fish=Fish(fishes=0)))


@dataclasses.dataclass(frozen=True)
class AppInherit(Bears, Fish):
    pass


class AppStateInherit(BearReactive[AppInherit], FishReactive[AppInherit], Reactive[AppInherit]):
    def eat(self):
        state = self.get()
        bears = state.count
        self.update(fishes=max(0, state.fishes - bears))


def test_app_inherit():
    mock = unittest.mock.Mock()
    app = AppInherit(type="brown", count=1, fishes=4)
    app_store = AppStateInherit(app)
    app_store.subscribe(mock)

    assert app_store.get().count == 1
    assert app_store.get().fishes == 4

    mock.assert_not_called()
    app_store.eat()
    assert app_store.get().fishes == 3
    assert mock.call_count == 1
    mock.assert_called_with(AppInherit(type="brown", count=1, fishes=3))

    app_store.increase_population()
    mock.assert_called_with(AppInherit(type="brown", count=2, fishes=3))
    assert mock.call_count == 2
    assert app_store.get().count == 2
    assert app_store.get().count == 2
    assert app_store.get().fishes == 3

    app_store.eat()
    assert app_store.get().fishes == 1
    assert mock.call_count == 3
    mock.assert_called_with(AppInherit(type="brown", count=2, fishes=1))
    app_store.eat()
    assert app_store.get().fishes == 0
    assert mock.call_count == 4
    mock.assert_called_with(AppInherit(type="brown", count=2, fishes=0))


def test_use_external_store():
    state = {"a": 1, "b": 2, "c": 3}
    update = lambda: None  # noqa

    def subscribe(callback: Callable[[], None]) -> Callable[[], None]:
        nonlocal update
        update = callback

        def cleanup():
            nonlocal update
            update = lambda: None  # noqa

        return cleanup

    renders = 0
    state_last = {}

    @react.component
    def Test():
        nonlocal renders
        nonlocal state_last
        state_last = use_sync_external_store(subscribe, lambda: state)
        renders += 1
        return sol.Button()

    box, rc = react.render(Test(), handle_error=False)
    assert renders == 1
    update()
    assert renders == 1
    assert state_last == state
    state = state.copy()
    state["b"] = 12
    update()
    assert state_last == state
    assert renders == 2


# TODO:
# def test_store_scope_application():
#     import solara as sol
#     import solara.scope

#     bear_store = BearReactive({"type": "brown", "count": 1}, storage=sol.scope.application)
#     try:
#         assert bear_store.get() == {"type": "brown", "count": 1}

#         @react.component
#         def BearCounter():
#             bears = bear_store.use(lambda bear: bear.count)
#             # bears = app_store.use(lambda app: app.bear.count)
#             return sol.Info(f"{bears} bears around here")

#         @react.component
#         def Controls():
#             return sol.Button("add bear", on_click=bear_store.increase_population)

#         @react.component
#         def App():
#             with sol.VBox() as main:
#                 BearCounter()
#                 Controls()
#             return main

#         app = App()
#         box, rc = react.render(app, handle_error=False)

#         assert rc._find(v.Alert).widget.children[0] == "1 bears around here"
#         bear_store.increase_population()
#         assert rc._find(v.Alert).widget.children[0] == "2 bears around here"

#         storage_key = bear_store.storage_key
#         redis_dict = bear_store._storage.observable_dict
#         redis_dict[storage_key] = {"type": "brown", "count": 3}
#         # TODO: wait for the store to update
#         time.sleep(0.1)
#         # assert rc._find(v.Alert).wait_for_widget.children[0] == "3 bears around here"
#         # assert rc._find(v.Alert).wait_for_single().widget.children[0] == "3 bears around here"
#         rc._find(v.Alert).widget.children[0] == "3 bears around here"
#     finally:
#         bear_store._storage.delete()
def test_dataframe():
    mock = unittest.mock.Mock()
    import pandas as pd

    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    df2 = df.copy()
    store = Reactive[pd.DataFrame](df)
    unsub = store.subscribe(mock)
    mock.assert_not_called()
    store.set(df)
    mock.assert_not_called()
    store.set(df2)
    mock.assert_called_with(df2)

    unsub()
    store.set(df)

    @solara.component
    def Test():
        df = store.use_value()
        return solara.Info(repr(id(df)))

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Alert).widget.children[0] == repr(id(df))
    df2 = df.copy()
    store.set(df2)
    assert rc.find(v.Alert).widget.children[0] == repr(id(df2))

    @dataclasses.dataclass(frozen=True)
    class DataFrame:
        df: Optional[pd.DataFrame] = dataclasses.field()

    df_store = Reactive[DataFrame](DataFrame(df=df))
    mock.reset_mock()
    unsub = Ref(df_store.fields).subscribe(mock)
    mock.assert_not_called()
    df_store.set(DataFrame(df=df2))
    mock.assert_called_once()
    mock.reset_mock()
    Ref(df_store.fields.df).set(df)
    mock.assert_called_once()
    unsub()



from typing import Callable

import reacton
import reacton.ipyvuetify as v


@reacton.value_component(bool)
def Checkbox(
    *,
    label=None,
    value=True,
    on_value: Callable[[bool], None] = None,
    disabled=False,
    style: str = None,
):
    """A checkbox is a widget that allows the user to toggle a boolean state.

    ## Arguments

     * `label`: The label to display next to the checkbox.
     * `value`: The current value of the checkbox (True or False).
     * `on_value`: A callback that is called when the checkbox is toggled.
     * `disabled`: If True, the checkbox is disabled and cannot be used.
     * `style`: A string of CSS styles to apply to the checkbox.
    """
    children = []
    if label is not None:
        children = [label]
    return v.Checkbox(label=label, v_model=value, on_v_model=on_value, disabled=disabled, style_=style, children=children)



import contextlib
import errno
import importlib
import inspect
import logging
import os
import sys
import threading
from typing import Callable, Dict, List, Optional, Set, Type

NO_WATCHDOG = False
try:
    from watchdog.events import FileSystemEventHandler
    from watchdog.observers import Observer
except ImportError:
    pass
    NO_WATCHDOG = True
logger = logging.getLogger("solara.server.reload")


class Watcher:
    def __init__(self, files, on_change: Callable[[str], None]):
        pass

    def close(self):
        pass

    def add_file(self, file):
        pass


class WatcherDummy:
    def __init__(self, files, on_change: Callable[[str], None]):
        pass

    def close(self):
        pass

    def add_file(self, file):
        pass


if NO_WATCHDOG:
    WatcherType: Type[Watcher] = WatcherDummy  # type: ignore

else:

    class WatcherWatchdog(FileSystemEventHandler):
        def __init__(self, files, on_change: Callable[[str], None]):
            self.on_change = on_change
            self.observers: List[Observer] = []
            self.directories: Set[str] = set()
            self.files: List[str] = []
            self.mtimes: Dict[str, float] = dict()
            for file in files:
                self.add_file(file)

        def close(self):
            for o in self.observers:
                o.unschedule_all()
                o.stop()
            # this makes testing slow
            # for o in self.observers:
            #     o.join()

        def add_file(self, file):
            file = os.path.abspath(os.path.realpath(file))
            if file not in self.files:
                logger.info("Watching file %s", file)
                if file in self.files:
                    raise RuntimeError(f"{file} was already added")
                self.files.append(file)
                self._watch_file(file)

        def _watch_file(self, file):
            self.mtimes[file] = os.path.getmtime(file)
            directory = os.path.realpath(os.path.dirname(file))
            self.watch_directory(directory)

        def watch_directory(self, directory):
            if directory not in self.directories:
                logger.debug("Watching directory %s", directory)
                observer = Observer()
                observer.schedule(self, directory, recursive=False)
                try:
                    observer.start()
                except OSError as e:
                    if e.errno == errno.ENOSPC:
                        # from rich import print()
                        from .utils import start_error

                        start_error(
                            "inotify watch limit reached", "Read https://github.com/widgetti/solara/#how-to-fix-inotify-watch-limit-reached", exception=e
                        )

                self.observers.append(observer)
                self.directories.add(directory)

        def on_modified(self, event):
            super(WatcherWatchdog, self).on_modified(event)
            logger.debug("Watch event: %s", event)
            if not event.is_directory:
                if event.src_path in self.files:
                    mtime_new = os.path.getmtime(event.src_path)
                    changed = mtime_new > self.mtimes[event.src_path]
                    self.mtimes[event.src_path] = mtime_new
                    if changed:
                        logger.debug("File modified: %s", event.src_path)
                        try:
                            self.on_change(event.src_path)
                        except:  # noqa
                            # we are in the watchdog thread here, all we can do is report
                            # and continue running (otherwise reload stops working)
                            logger.exception("Error while executing on change handler")
                    else:
                        logger.debug("File reported modified, but mtime did not change: %s", event.src_path)
                else:
                    logger.debug("Ignore file modification: %s", event.src_path)

    WatcherType = WatcherWatchdog  # type: ignore


class Reloader:
    def __init__(self, on_change: Optional[Callable[[str], None]] = None) -> None:
        self.watched_modules: Set[str] = set()
        self._first = True
        self.on_change: Optional[Callable[[str], None]] = on_change
        self.watcher = WatcherType([], self._on_change)
        self.requires_reload = False
        self.ignore_modules: Set[str] = set()
        self.reload_event_next = threading.Event()

    def start(self):
        if self._first:
            # although we try to keep track of modules loaded (which we will watch)
            # it can happen that an import is done at runtime, that we miss (could be in a thread)
            # so we always reload all modules except the ignore_modules
            self.ignore_modules = set(sys.modules)
            logger.info("Ignoring reloading modules: %s", self.ignore_modules)
            self._first = False

    def _on_change(self, name):
        # used for testing
        self.reload_event_next.set()
        self.reload_event_next.clear()
        # flag that we need to reload all modules next time
        self.requires_reload = True
        # and forward callback
        if self.on_change:
            self.on_change(name)

    def close(self):
        self.watcher.close()

    def reload(self):
        # before we did this:
        # # don't reload modules like solara.server and react
        # # that may cause issues (like 2 Element classes existing)
        # not sure why, but if we reload pandas, the integration/reload_test.py fails
        reload_modules = {
            k for k in set(sys.modules) - set(self.ignore_modules) if not (k.startswith("solara.server") or k.startswith("anyio") or k.startswith("pandas"))
        }
        # which picks up import that are done in threads etc, but it will also reload starlette, httptools etc
        # which causes issues with exceptions and isinstance checks.
        # reload_modules = self.watched_modules
        logger.info("Reloading modules... %s", reload_modules)
        # not sure if this is needed
        importlib.invalidate_caches()
        for mod in sorted(reload_modules):
            # don't reload modules like solara.server and react
            # that may cause issues (like 2 Element classes existing)
            logger.debug("Reloading module %s", mod)
            sys.modules.pop(mod, None)
        # if all succesfull...
        self.requires_reload = False

    @contextlib.contextmanager
    def watch(self):
        """Use this context manager to execute code so that we can track loaded modules and reload them when needed"""
        if self.requires_reload:
            self.reload()

        modules_before = set(sys.modules)
        try:
            yield
        finally:
            modules_after = set(sys.modules)
            # these are imported during the 'yield'
            modules_added = modules_after - modules_before
            modules_to_consider = modules_added - set(self.watched_modules)
            # TODO: libraries that solara uses need special care
            # modules_always = {k for k in sys.modules if k.startswith("reacton")}
            # modules_to_consider = modules_to_consider | modules_always
            modules_watching = set()
            if modules_to_consider:
                logger.debug("Found modules %s", modules_to_consider)
            for modname in modules_to_consider:
                module = sys.modules[modname]
                path = None
                try:
                    path = inspect.getfile(module)
                except Exception:
                    pass
                if path:
                    if not path.startswith(sys.prefix):
                        self.watcher.add_file(path)
                        self.watched_modules.add(modname)
                        modules_watching.add(modname)
            if modules_watching:
                logger.info("Watching modules: %s for reload", modules_watching)


# there is only a reloader, and there should be only 1 app
# that connect to the on_change
reloader = Reloader()



import logging
import os
import sys
import typing
from typing import List, Union, cast
from uuid import uuid4

import anyio
import starlette.websockets
import uvicorn.server
import websockets.legacy.http
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles

import solara
from solara.server.threaded import ServerBase

from . import app as appmod
from . import server, settings, telemetry, websocket
from .cdn_helper import cdn_url_path, get_path

os.environ["SERVER_SOFTWARE"] = "solara/" + str(solara.__version__)

logger = logging.getLogger("solara.server.fastapi")
# if we add these to the router, the server_test does not run (404's)
prefix = ""

# The limit for starlette's http traffic should come from h11's DEFAULT_MAX_INCOMPLETE_EVENT_SIZE=16kb
# In practice, testing with 132kb cookies (server_test.py:test_large_cookie) seems to work fine.
# For the websocket, the limit is set to 4kb till 10.4, see
#  * https://github.com/aaugustin/websockets/blob/10.4/src/websockets/legacy/http.py#L14
# Later releases should set this to 8kb. See
#  * https://github.com/aaugustin/websockets/commit/8ce4739b7efed3ac78b287da7fb5e537f78e72aa
#  * https://github.com/aaugustin/websockets/issues/743
# Since starlette seems to accept really large values for http, lets do the same for websockets
# An arbitrarily large value we settled on for now is 32kb
# If we don't do this, users with many cookies will fail to get a websocket connection.
websockets.legacy.http.MAX_LINE = 1024 * 32


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: starlette.websockets.WebSocket

    def __init__(self, ws: starlette.websockets.WebSocket, portal: anyio.from_thread.BlockingPortal) -> None:
        self.ws = ws
        self.portal = portal

    def close(self):
        self.portal.call(self.close)

    def send_text(self, data: str) -> None:
        self.portal.call(self.ws.send_text, data)

    def send_bytes(self, data: bytes) -> None:
        self.portal.call(self.ws.send_bytes, data)

    def receive(self):
        message = self.portal.call(self.ws.receive)
        if "text" in message:
            return message["text"]
        elif "bytes" in message:
            return message["bytes"]
        elif message.get("type") == "websocket.disconnect":
            raise websocket.WebSocketDisconnect()
        else:
            raise RuntimeError(f"Unknown message type {message}")


class ServerStarlette(ServerBase):
    server: uvicorn.server.Server

    def has_started(self):
        return self.server.started

    def signal_stop(self):
        self.server.should_exit = True
        # this cause uvicorn to not wait for background tasks, e.g.:
        # <Task pending name='Task-55'
        #  coro=<WebSocketProtocol.run_asgi() running at
        #  /.../uvicorn/protocols/websockets/websockets_impl.py:184>
        # wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x16896aa00>()]>
        # cb=[WebSocketProtocol.on_task_complete()]>
        self.server.force_exit = True
        self.server.lifespan.should_exit = True

    def serve(self):

        from uvicorn.config import Config
        from uvicorn.server import Server

        if sys.version_info[:2] < (3, 7):
            # make python 3.6 work
            import asyncio

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # uvloop will trigger a: RuntimeError: There is no current event loop in thread 'fastapi-thread'
        config = Config(app, host=self.host, port=self.port, **self.kwargs, loop="asyncio")
        self.server = Server(config=config)
        self.started.set()
        self.server.run()


async def kernels(id):
    return JSONResponse({"name": "lala", "id": "dsa"})


async def kernel_connection(ws: starlette.websockets.WebSocket):
    session_id = ws.cookies.get(server.COOKIE_KEY_SESSION_ID)
    if not session_id:
        logger.error("no session cookie")
        await ws.close()
        return
    connection_id = ws.query_params["session_id"]
    if not connection_id:
        logger.error("no session_id/connection_id")
        await ws.close()
        return
    logger.info("Solara kernel requested for session_id=%s connection_id=%s", session_id, connection_id)
    await ws.accept()

    def websocket_thread_runner(ws: starlette.websockets.WebSocket, portal: anyio.from_thread.BlockingPortal):
        ws_wrapper = WebsocketWrapper(ws, portal)

        async def run():
            try:
                assert session_id is not None
                assert connection_id is not None
                await server.app_loop(ws_wrapper, session_id, connection_id)
            except:  # noqa
                await portal.stop(cancel_remaining=True)
                raise

        # sometimes throws: RuntimeError: Already running asyncio in this thread
        anyio.run(run)

    # this portal allows us to sync call the websocket calls from this current event loop we are in
    # each websocket however, is handled from a separate thread
    try:
        async with anyio.from_thread.BlockingPortal() as portal:
            thread_return = anyio.to_thread.run_sync(websocket_thread_runner, ws, portal)
            await thread_return
    finally:
        try:
            await ws.close()
        except:  # noqa
            pass


async def close(request: Request):
    connection_id = request.path_params["connection_id"]
    if connection_id in appmod.contexts:
        context = appmod.contexts[connection_id]
        context.close()
    response = HTMLResponse(content="", status_code=200)
    return response


async def root(request: Request, fullpath: str = ""):
    root_path = settings.main.root_path or ""
    # if not explicltly set,
    if settings.main.root_path is None:
        # use the default root path from the app, which seems to also include the path
        # if we are mounted under a path
        root_path = request.scope.get("root_path", "")
        logger.debug("root_path: %s", root_path)
        # or use the script-name header, for instance when running under a reverse proxy
        script_name = request.headers.get("script-name")
        if script_name:
            logger.debug("override root_path using script-name header from %s to %s", root_path, script_name)
            root_path = script_name
        script_name = request.headers.get("x-script-name")
        if script_name:
            logger.debug("override root_path using x-script-name header from %s to %s", root_path, script_name)
            root_path = script_name
        settings.main.root_path = root_path

    request_path = request.url.path
    if request_path.startswith(root_path):
        request_path = request_path[len(root_path) :]
    content = server.read_root(request_path, root_path)
    if content is None:
        return HTMLResponse(content="Page not found by Solara router", status_code=404)
    response = HTMLResponse(content=content)
    session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID) or str(uuid4())
    samesite = "lax"
    secure = False
    # we want samesite, so we can set a cookie when embedded in an iframe, such as on huggingface
    # however, samesite=none requires Secure https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite
    if request.headers.get("x-forwarded-proto", "http") == "https":
        samesite = "none"
        secure = True
    response.set_cookie(
        server.COOKIE_KEY_SESSION_ID, value=session_id, expires="Fri, 01 Jan 2038 00:00:00 GMT", samesite=samesite, secure=secure  # type: ignore
    )  # type: ignore
    return response


class StaticNbFiles(StaticFiles):
    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        return cast(List[Union[str, "os.PathLike[str]"]], server.nbextensions_directories)

    # follow symlinks
    # from https://github.com/encode/starlette/pull/1377/files
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        for directory in self.all_directories:
            original_path = os.path.join(directory, path)
            full_path = os.path.realpath(original_path)
            directory = os.path.realpath(directory)
            try:
                return full_path, os.stat(full_path)
            except (FileNotFoundError, NotADirectoryError):
                continue
        return "", None


class StaticPublic(StaticFiles):
    def lookup_path(self, *args, **kwargs):
        self.all_directories = self.get_directories(None, None)
        return super().lookup_path(*args, **kwargs)

    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        # we only know the .directory at runtime (after startup)
        # which means we cannot pass the directory to the StaticFiles constructor
        return cast(List[Union[str, "os.PathLike[str]"]], [app.directory.parent / "public" for app in appmod.apps.values()])


class StaticAssets(StaticFiles):
    def lookup_path(self, *args, **kwargs):
        self.all_directories = self.get_directories(None, None)
        return super().lookup_path(*args, **kwargs)

    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        # we only know the .directory at runtime (after startup)
        # which means we cannot pass the directory to the StaticFiles constructor
        overrides = [app.directory.parent / "assets" for app in appmod.apps.values()]
        default = server.solara_static.parent / "assets"
        return cast(List[Union[str, "os.PathLike[str]"]], [*overrides, default])


class StaticCdn(StaticFiles):
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        full_path = str(get_path(settings.assets.proxy_cache_dir, path))
        return full_path, os.stat(full_path)


def on_startup():
    # TODO: configure and set max number of threads
    # see https://github.com/encode/starlette/issues/1724
    telemetry.server_start()


def on_shutdown():
    telemetry.server_stop()


def readyz(request: Request):
    json, status = server.readyz()
    return JSONResponse(json, status_code=status)


middleware = [Middleware(GZipMiddleware, minimum_size=1000)]

routes = [
    Route("/readyz", endpoint=readyz),
    Route("/jupyter/api/kernels/{id}", endpoint=kernels),
    WebSocketRoute("/jupyter/api/kernels/{id}/{name}", endpoint=kernel_connection),
    Route("/", endpoint=root),
    Route("/{fullpath}", endpoint=root),
    Route("/_solara/api/close/{connection_id}", endpoint=close, methods=["POST"]),
    Mount(f"/{cdn_url_path}", app=StaticCdn(directory=settings.assets.proxy_cache_dir)),
    Mount(f"{prefix}/static/public", app=StaticPublic()),
    Mount(f"{prefix}/static/assets", app=StaticAssets()),
    Mount(f"{prefix}/static/nbextensions", app=StaticNbFiles()),
    Mount(f"{prefix}/static/nbconvert", app=StaticFiles(directory=server.nbconvert_static)),
    Mount(f"{prefix}/static", app=StaticFiles(directory=server.solara_static)),
    Route("/{fullpath:path}", endpoint=root),
]

app = Starlette(routes=routes, on_startup=[on_startup], on_shutdown=[on_shutdown], middleware=middleware)
# Uncomment the lines below to test solara mouted under a subpath
# def myroot(request: Request):
#     return JSONResponse({"framework": "solara"})

# routes_test_sub = [Route("/", endpoint=myroot), Mount("/foo/", routes=routes)]
# app = Starlette(routes=routes_test_sub, on_startup=[on_startup], on_shutdown=[on_shutdown], middleware=middleware)



import os
from typing import Optional

import pydantic
from pydantic import Field

from .util import get_solara_home

home = get_solara_home()
if not home.exists():
    home.mkdir(parents=True, exist_ok=True)


class Cache(pydantic.BaseSettings):
    type: str = pydantic.Field("memory", env="SOLARA_CACHE", title="Type of cache, e.g. 'memory', 'disk', 'redis', or a multilevel cache, e.g. 'memory,disk'")
    disk_max_size: str = Field("10GB", title="Maximum size for'disk' cache , e.g. 10GB, 500MB")
    memory_max_size: str = Field("1GB", title="Maximum size for 'memory-size' cache, e.g. 10GB, 500MB")
    memory_max_items: int = Field(128, title="Maximum number of items for 'memory' cache")
    clear: bool = Field(False, title="Clear the cache on startup, only applies to disk and redis caches")
    path: Optional[str] = Field(
        os.path.join(home, "cache"), env="SOLARA_CACHE_PATH", title="Storage location for 'disk' cache. Defaults to `${SOLARA_HOME}/cache`"
    )

    class Config:
        env_prefix = "solara_cache_"
        case_sensitive = False
        env_file = ".env"


cache: Cache = Cache()
